/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "meshM2.h"

#include "graphics/igraphics.h"
#include "graphics/isampler.h"
#include "graphics/ishader.h"

#include "graphics/iindexbuffer.h"
#include "graphics/ivertexbuffer.h"

#include "kernel.h"

namespace hydra
{
IMPLEMENT_MODULE(MeshM2, usr/lib/graphics/mesh.m2)

enum {
	SIMPLE = 0,
	REFLECT,
	MAX_SHADER,
};

#if defined(_PS3) && !defined(_NVCG)
	extern uint_t _binary_simple_vp_vpo_start;
	extern uint_t _binary_simple_vp_vpo_end;

	extern uint_t _binary_simple_fp_fpo_start;
	extern uint_t _binary_simple_fp_fpo_end;

	extern uint_t _binary_reflect_vp_vpo_start;
	extern uint_t _binary_reflect_vp_vpo_end;

	extern uint_t _binary_reflect_fp_fpo_start;
	extern uint_t _binary_reflect_fp_fpo_end;

	static char * vpshaders[MAX_SHADER];
	static char * fpshaders[MAX_SHADER];
#endif //_PS3

static const char * shaders[MAX_SHADER] = {
	"shaders/simple.cg",
	"shaders/reflect.cg",
};

MeshM2::MeshM2()
{
#if defined(_PS3) && !defined(_NVCG)
	vpshaders[0] = (char *) &_binary_simple_vp_vpo_start;
	fpshaders[0] = (char *) &_binary_simple_fp_fpo_start;

	vpshaders[1] = (char *) &_binary_reflect_vp_vpo_start;
	fpshaders[1] = (char *) &_binary_reflect_fp_fpo_start;
#endif //_PS3
}

typedef unsigned char	uchar;
typedef unsigned short	ushort;
typedef unsigned int	uint;
typedef unsigned long	ulong;

#pragma pack(1)
struct M2Header {
	char id[4];
	uchar version[4];
	uint nameLength;
	uint nameOfs;
	uint type;

	uint nGlobalSequences;
	uint ofsGlobalSequences;
	uint nAnimations;
	uint ofsAnimations;
	uint nC;
	uint ofsC;
	uint nD;
	uint ofsD;
	uint nBones;
	uint ofsBones;
	uint nF;
	uint ofsF;

	uint nVertices;
	uint ofsVertices;
	uint nViews;
	uint ofsViews;

	uint nColors;
	uint ofsColors;

	uint nTextures;
	uint ofsTextures;

	uint nTransparency; // H
	uint ofsTransparency;
	uint nI;   // always unused ?
	uint ofsI;
	uint nTexAnims;	// J
	uint ofsTexAnims;
	uint nTexReplace;
	uint ofsTexReplace;

	uint nTexFlags;
	uint ofsTexFlags;
	uint nY;
	uint ofsY;

	uint nTexLookup;
	uint ofsTexLookup;

	uint nTexUnitLookup;		// L
	uint ofsTexUnitLookup;
	uint nTransparencyLookup; // M
	uint ofsTransparencyLookup;
	uint nTexAnimLookup;
	uint ofsTexAnimLookup;

	float floats[14];

	uint nBoundingTriangles;
	uint ofsBoundingTriangles;
	uint nBoundingVertices;
	uint ofsBoundingVertices;
	uint nBoundingNormals;
	uint ofsBoundingNormals;

	uint nAttachments; // O
	uint ofsAttachments;
	uint nAttachLookup; // P
	uint ofsAttachLookup;
	uint nQ; // Q
	uint ofsQ;
	uint nLights; // R
	uint ofsLights;
	uint nCameras; // S
	uint ofsCameras;
	uint nT;
	uint ofsT;
	uint nRibbonEmitters; // U
	uint ofsRibbonEmitters;
	uint nParticleEmitters; // V
	uint ofsParticleEmitters;
};

// block B - animations
struct M2Animation {
	uint animID;
	uint timeStart;
	uint timeEnd;

	float moveSpeed;

	uint loopType;
	uint flags;
	uint d1;
	uint d2;
	uint playSpeed;  // note: this can't be play speed because it's 0 for some models

	float3 boxA, boxB;
	float rad;

	ushort s[2];
};


// sub-block in block E - animation data
struct AnimationBlock {
	ushort type;		// interpolation type (0=none, 1=linear, 2=hermite)
	ushort seq;		// global sequence id or -1
	uint nRanges;
	uint ofsRanges;
	uint nTimes;
	uint ofsTimes;
	uint nKeys;
	uint ofsKeys;
};

// block E - bones
struct M2BoneDef {
	uint animid;
	uint flags;
	ushort parent; // parent bone index
	ushort geoid;
	// new int added to the bone definitions.  Added in WoW 2.0
	uint unknown;
	AnimationBlock translation;
	AnimationBlock rotation;
	AnimationBlock scaling;
	float3 pivot;
};

struct M2TexAnimDef {
	AnimationBlock trans, rot, scale;
};

struct M2Vertex {
	float3 pos;
	uchar weights[4];
	uchar bones[4];
	float3 normal;
	float2 texcoords;
	int unk1, unk2; // always 0,0 so this is probably unused
};

struct M2View {
	uint nIndex, ofsIndex; // Vertices in this model (index into vertices[])
	uint nTris, ofsTris;	 // indices
	uint nProps, ofsProps; // additional vtx properties
	uint nSub, ofsSub;	 // materials/renderops/submeshes
	uint nTex, ofsTex;	 // material properties/textures
	int lod;				 // LOD bias?
};

/// One material + render operation
struct M2Geoset {
	ushort id;		// mesh part id?
	ushort d2;		// ?
	ushort vstart;	// first vertex
	ushort vcount;	// num vertices
	ushort istart;	// first index
	ushort icount;	// num indices
	ushort d3;		// number of bone indices
	ushort d4;		// ? always 1 to 4
	ushort d5;		// ?
	ushort d6;		// root bone?
	float3 v;
	float unknown[4];	// Added in WoW 2.0?
};

struct M2TextureDef {
	uint type;
	uint flags;
	uint nameLen;
	uint nameOfs;
};

/// A texture unit (sub of material)
struct M2TexUnit{
	// probably the texture units
	// size always >=number of materials it seems
	ushort flags;		// Flags
	ushort order;		// ?
	ushort op;			// Material this texture is part of (index into mat)
	ushort op2;			// Always same as above?
	short colorIndex;	// color or -1
	ushort flagsIndex;	// more flags...
	ushort texunit;		// Texture unit (0 or 1)
	ushort d4;			// ? (seems to be always 1)
	ushort textureid;	// Texture id (index into global texture list)
	ushort texunit2;	// copy of texture unit value?
	ushort transid;		// transparency id (index into transparency list)
	ushort texanimid;	// texture animation id
};

// block X - render flags
struct M2RenderFlags {
	ushort flags;
	//unsigned char f1;
	//unsigned char f2;
	ushort blend;
};

// block G - color defs
struct M2ColorDef {
	AnimationBlock color;
	AnimationBlock opacity;
};

// block H - transp defs
struct M2TransDef {
	AnimationBlock trans;
};

struct M2LightDef {
	ushort type;
	ushort bone;
	float3 pos;
	AnimationBlock ambColor;
	AnimationBlock ambIntensity;
	AnimationBlock color;
	AnimationBlock intensity;
	AnimationBlock attStart;
	AnimationBlock attEnd;
	AnimationBlock unk1;
};

struct M2CameraDef {
	uint id;
	float fov, farclip, nearclip;
	AnimationBlock transPos;
	float3 pos;
	AnimationBlock transTarget;
	float3 target;
	AnimationBlock rot;
};


struct M2ParticleParams {
	float mid;
	uint colors[3];
	float sizes[3];
	ushort d[10];
	float unk[3];
	float scales[3];
	float slowdown;
	float rotation;	//Sprite Rotation
	float unknown;
	float Rot1[3];	//Model Rotation 1
	float Rot2[3];	//Model Rotation 2
	float Trans[3];	//Model Translation
	float f2[6];
};

struct M2ParticleEmitterDef {
	uint id;
	uint flags;
	float3 pos;
	ushort bone;
	ushort texture;
	uint nModelName;
	uint ofsModelName;
	uint nZero2;
	uint ofsZero2;
	ushort blend;
	ushort type;
	ushort s1;
	ushort s2;
	ushort cols;
	ushort rows;
	AnimationBlock params[10];
	M2ParticleParams p;
	AnimationBlock en;
};

struct M2RibbonEmitterDef {
	uint id;
	uint bone;
	float3 pos;
	uint nTextures;
	uint ofsTextures;
	uint nUnknown;
	uint ofsUnknown;
	AnimationBlock color;
	AnimationBlock opacity;
	AnimationBlock above;
	AnimationBlock below;
	float res, length, unk;
	ushort s1, s2;
	AnimationBlock unk1;
	AnimationBlock unk2;
};

struct M2BlockQ {
	char id[4];
	uint dbid;
	uint bone;
	float3 pos;
	ushort type;
	ushort seq;
	uint nRanges;
	uint ofsRanges;
	uint nTimes;
	uint ofsTimes;
};

struct M2AttachmentDef {
	uint id;
	uint bone;
	float3 pos;
	AnimationBlock unk;
};

struct M2RenderPass {
	uint indexStart, indexCount, vertexStart, vertexEnd;

	//TextureID texture, texture2;
	int tex;
	bool useTex2, useEnvMap, cull, trans, unlit, noZWrite, billboard;
	float p;

	short texanim, color, opacity, blendmode;
	ushort order;

	// Geoset ID
	int geoset;

	// texture wrapping
	bool swrap, twrap;

	// colours
	float4 ocol, ecol;
};
#pragma pack()

static ISampler * loadTexture(char * buffer, int texId)
{
	ISampler * sampler = NULL;

	//
	M2Header * header = (M2Header *) buffer;
	M2TextureDef * texdef = (M2TextureDef *)(buffer + swapEndian(header->ofsTextures));

	int nTextures = swapEndian(header->nTextures);
	if (nTextures == 0)
		return NULL;

	char texname[MAX_PATH];
	int type = swapEndian(texdef[texId].type);
	if (type == 0)
	{
		int nameLen = swapEndian(texdef[texId].nameLen);
		strncpy(texname, (const char *) buffer + swapEndian(texdef[texId].nameOfs), nameLen);
		texname[nameLen] = 0;

		// change ext
		char * ptr = (char *) _strchr(texname, '.');
		strcpy(ptr, ".DDS");

		// trim folder name
		ptr = (char *) texname + (strlen(texname) - 1);
		while (*ptr)
		{
			if (*ptr == '\\' || *ptr == '/')
			{
				ptr ++;
				break;
			}

			ptr --;
		}

		char fullPath[MAX_PATH];
		strcpy(fullPath, "textures/");
		strcat(fullPath, ptr);

		TRACE(">> TEX(%d): %s\n", texId, fullPath)
		sampler = loadSampler(fullPath);
	}
	else
	{
		if (type == 3)
		{
			//TRACE("# TEX(%d): Item\\ObjectComponents\\Weapon\\ArmorReflect4.BLP",texId)

			// a fix for weapons with type-3 textures.
			const char * reflect = "textures/ARMORREFLECT4.DDS";
			sampler = loadSampler(reflect);
		}
		else
		{
			char name[MAX_PATH];
			char * nameOfs = (char *) header + swapEndian(header->nameOfs);
			strncpy(name, nameOfs, swapEndian(header->nameLength));

            if (type == 12)
            {
                name[swapEndian(header->nameLength) - 2] = 0;
                sprintf(texname, "textures/%sArmor.DDS", name);
            }
            else //11
                sprintf(texname, "textures/%sSkin.DDS", name);

			TRACE(">> TEX(%d): %s, %d\n", texId, texname, type)
			sampler = loadSampler(texname);
		}
	}

    //
	if (!sampler)
	{
	    const char * dust = "textures/DUST2.DDS";
		TRACE(">> TEX(%d): %s\n", texId, dust)
		sampler = loadSampler(dust);
	}

	return sampler;
}

bool MeshM2::loadFromFile(const char * pathName)
{
	char fullPath[MAX_PATH];
	makeFullPath(fullPath, pathName);

	size_t length;
	char * buffer = (char *) ::loadFromFile(fullPath, &length);
	if (!buffer)
		return false;

	M2Header * header = (M2Header *) buffer;
	if (header->id[0] != 'M' && header->id[1] != 'D' && header->id[2] != '2' && header->id[3] != '0')
	{
		TRACE0(">> ERROR: Invalid model!  May be corrupted.\n")

		SAFE_FREE(buffer);
		return false;
	}

	// 4 1 0 0 = WOW 2.0 models
	// 0 1 0 0 = WOW 1.0 models
	if (header->version[0] != 4 && header->version[1] != 1 && header->version[2] != 0 && header->version[3] != 0)
	{
		TRACE0(">> ERROR: Model version is incorrect!\nMake sure you are loading models from World of Warcraft 2.0.1 or newer client.\n")

		if (header->version[0] == 0)
		{
			TRACE("An error occured while trying to load the model %s.\nWoW Model Viewer 0.5.x only supports loading WoW 2.0 models\nModels from WoW 1.12 or earlier are not supported.\n", pathName)
		}

		SAFE_FREE(buffer);
		return false;
	}

	if (length < swapEndian(header->ofsParticleEmitters))
	{
		TRACE0(">> ERROR: Unable to load the Model \"%s\", appears to be corrupted.\n", pathName)

		SAFE_FREE(buffer);
		return false;
	}

	char name[MAX_PATH];
	char * nameOfs = (char *) header + swapEndian(header->nameOfs);
	strncpy(name, nameOfs, swapEndian(header->nameLength));
	//setName(name);

	TRACE(">> name: %s\n", name)

	if (swapEndian(header->nGlobalSequences))
	{
		//TRACE(">> GlobalSequences: %d", swapEndian(header->nGlobalSequences))

		//globalSequences = new int[header.nGlobalSequences];
		//memcpy(globalSequences, (f.getBuffer() + header.ofsGlobalSequences), header.nGlobalSequences * 4);
	}

	//
	unsigned long fvf = FVF_XYZB5 | FVF_LASTBETA_UBYTE4 | FVF_NORMAL | FVF_TEX1;
	m_vertexbuffer = createVertexbuffer(swapEndian(header->nVertices), fvf);
	if (!m_vertexbuffer)
	{
		SAFE_FREE(buffer);
		return false;
	}
	
	//
	M2Vertex * vertices = (M2Vertex *)(buffer + swapEndian(header->ofsVertices));

	char * vertexBuffer = (char *) m_vertexbuffer->map();
	for (int n=0; n<(int) swapEndian(header->nVertices); n++)
	{
		float3 * position = (float3 *) vertexBuffer;
		position->x = -swapEndian(vertices->pos.x);
		position->y =  swapEndian(vertices->pos.z);
		position->z = -swapEndian(vertices->pos.y);
		vertexBuffer += sizeof(float3);

		float4 * weights = (float4 *) vertexBuffer;
		weights->x = (float) vertices->weights[0];
		weights->y = (float) vertices->weights[1];
		weights->z = (float) vertices->weights[2];
		weights->w = (float) vertices->weights[3];
		vertexBuffer += sizeof(float4);

		unsigned char * indices = (unsigned char *) vertexBuffer;
		indices[0] = vertices->bones[0];
		indices[1] = vertices->bones[1];
		indices[2] = vertices->bones[2];
		indices[3] = vertices->bones[3];
		vertexBuffer += sizeof(uchar) * 4;

		float3 * normal = (float3 *) vertexBuffer;
		normal->x = -swapEndian(vertices->normal.x);
		normal->y =  swapEndian(vertices->normal.z);
		normal->z = -swapEndian(vertices->normal.y);
		vertexBuffer += sizeof(float3);

		float2 * texcoord = (float2 *) vertexBuffer;
		texcoord->x = swapEndian(vertices->texcoords.x);
		texcoord->y = swapEndian(vertices->texcoords.y);
		vertexBuffer += sizeof(float2);

		vertices ++;
	}
	m_vertexbuffer->unmap();

	// indices - allocate space, too
	M2View * view = (M2View *)(buffer + swapEndian(header->ofsViews));

	// Indices,  Triangles
	unsigned short * indexLookup = (unsigned short *)(buffer + swapEndian(view->ofsIndex));
	unsigned short * triangles = (unsigned short *)(buffer + swapEndian(view->ofsTris));

	m_indexbuffer = createIndexbuffer(swapEndian(view->nTris), FMT_INDEX16);
	if (!m_indexbuffer)
	{
		SAFE_FREE(buffer);
		return false;
	}

	char * indexBuffer = (char *) m_indexbuffer->map();
	for (int i = 0; i<(int) swapEndian(view->nTris); i++)
	{
		unsigned short index = indexLookup[swapEndian(triangles[i])];

		unsigned short * v = (unsigned short *) indexBuffer;
		*v = swapEndian(index);
		indexBuffer += sizeof(short);
	}
	m_indexbuffer->unmap();

	// init colors
	if (swapEndian(header->nColors))
	{
		//TRACE(">> colors: %d", swapEndian(header->nColors));

		/*colors = new ModelColor[header.nColors];
		ModelColorDef *colorDefs = (ModelColorDef*)(f.getBuffer() + header.ofsColors);
		for (size_t i=0; i<header.nColors; i++)
			colors[i].init(f, colorDefs[i], globalSequences);
		*/
	}

	// init transparency
	short * transLookup = (short *)(buffer + swapEndian(header->ofsTransparencyLookup));
	if (swapEndian(header->nTransparency))
	{
		//TRACE(">> colors: %d", swapEndian(header->nTransparency))

		/*transparency = new ModelTransparency[header.nTransparency];
		ModelTransDef *trDefs = (ModelTransDef*)(f.getBuffer() + header.ofsTransparency);
		for (size_t i=0; i<header.nTransparency; i++)
			transparency[i].init(f, trDefs[i], globalSequences);
			*/
	}

	// render ops
	M2Geoset * ops = (M2Geoset*)(buffer + swapEndian(view->ofsSub));
	M2TexUnit * tex = (M2TexUnit*)(buffer + swapEndian(view->ofsTex));
	M2RenderFlags * renderFlags = (M2RenderFlags *)(buffer + swapEndian(header->ofsTexFlags));
	ushort * texlookup = (ushort *)(buffer + swapEndian(header->ofsTexLookup));
	ushort * texanimlookup = (ushort *)(buffer + swapEndian(header->ofsTexAnimLookup));
	ushort * texunitlookup = (ushort*)(buffer + swapEndian(header->ofsTexUnitLookup));

	int count = 0;
	int nTex = swapEndian(view->nTex);

	//TRACE(">> nTex: %d", nTex)

	// check envmap
	for (int k=0 ; k<nTex; k++)
	{
		int geoset = swapEndian(tex[k].op);
		M2RenderFlags &rf = renderFlags[swapEndian(tex[k].flagsIndex)];

		ushort lookup = texunitlookup[swapEndian(tex[k].texunit)];
		bool useEnvMap = (swapEndian(lookup) == 0xffff) && ((swapEndian(rf.flags) & 0x10) != 0) && (swapEndian(rf.blend)) > 2; //&& rf.blend<5; // Use environmental reflection effects?
		if (!useEnvMap)
			count ++;
	}

    //
    m_batch = new Batch [count + 1];
    if (!m_batch)
	{
		SAFE_FREE(buffer);
		return false;
	}

	Batch * batch = m_batch;
	Batch * envbatch = batch;

	count = 0;
	for (int j = 0; j<nTex; j++)
	{
		int geoset = swapEndian(tex[j].op);
		int texId = swapEndian(texlookup[swapEndian(tex[j].textureid)]);


		// Render Flags
		//0x01 = Unlit
		//0x02 = ? glow effects ? no zwrite?
		//0x04 = ?
		//0x08 = ?
		//0x10 = Billboard
		//0x20 = ?

		M2RenderFlags &rf = renderFlags[swapEndian(tex[j].flagsIndex)];

		ushort lookup = texunitlookup[swapEndian(tex[j].texunit)];
		bool useEnvMap = (swapEndian(lookup) == 0xffff) && ((swapEndian(rf.flags) & 0x10) != 0) && (swapEndian(rf.blend)) > 2; //&& rf.blend<5; // Use environmental reflection effects?
		if (useEnvMap)
		{
			IShader * shader = envbatch->shader;
			SAFE_RELEASE(shader);

            //
#if defined(_PS3) && !defined(_NVCG)
			shader = createShader(shaders[REFLECT]);
			if (!shader->loadFromMemory(vpshaders[REFLECT], fpshaders[REFLECT]))
            {
				SAFE_FREE(buffer);
                return false;
            }
//#endif //_PS3
#else
			shader = loadShader(shaders[REFLECT]);
			if (!shader)
            {
				SAFE_FREE(buffer);
                return false;
            }
#endif	//else

			envbatch->shader = shader;

			ISampler * sampler = loadTexture(buffer, texId);
			envbatch->sampler[2] = sampler;

			float4 specular(0.9f, 0.9f, 0.9f, 1.0f);
			envbatch->specular = specular;

			continue;
		}

		float4 ambient(0.4f, 0.4f, 0.4f, 1.0f);
		float4 diffuse(0.5f, 0.5f, 0.5f, 1.0f);
		float4 specular(0.3f, 0.3f, 0.3f, 1.0f);

		batch->ambient = ambient;
		batch->diffuse = diffuse;
		batch->specular = specular;

		batch->startIndex = swapEndian(ops[geoset].istart);
		batch->indexCount = swapEndian(ops[geoset].icount);

		//pass[j].order = tex[j].order; //pass.order = 0;
		//pass[j].tex = texlookup[tex[j].textureid];

		//pass[j].blendmode = rf.blend;
		//pass[j].color = tex[j].colorIndex;

		batch->flag = swapEndian(rf.blend);

        //
#if defined(_PS3) && !defined(_NVCG)
		IShader * shader = createShader(shaders[SIMPLE]);
		if (!shader->loadFromMemory(vpshaders[SIMPLE], fpshaders[SIMPLE]))
        {
			SAFE_FREE(buffer);
            return false;
        }
//#endif //_PS3
#else
		IShader * shader = loadShader(shaders[SIMPLE]);
        if (!shader)
		{
			SAFE_FREE(buffer);
            return false;
        }
#endif	//else

		batch->shader = shader;

		ISampler * sampler = NULL;
		if (!strcmp(name, "DragonNefarian"))
		{
			int ti[] = { 3, 1, 2, 1, 1, 3};
			char fileName[MAX_PATH];
			sprintf(fileName, "textures/DragonNefarian%d.DDS", ti[j]);

			TRACE(">> TEX(%d): %s\n", texId, fileName)
			sampler = loadSampler(fileName);
		}
		if (!strcmp(name, "DragonAzurgoz"))
		{
			int ti[] = { 3, 3, 2, 1, 3, 3, 1, 2, 1 };
			char fileName[MAX_PATH];
			sprintf(fileName, "textures/DRAGONSKIN%dRED.DDS", ti[j]);

			TRACE(">> TEX(%d): %s\n", texId, fileName)
			sampler = loadSampler(fileName);
		}

		//
		if (!sampler)
			sampler = loadTexture(buffer, texId);

		batch->sampler[0] = sampler;

		envbatch = batch;
		batch ++;
	}

	SAFE_FREE(buffer);

	TRACE("MeshM2::loadFromFile(%s)\n", pathName)
	
    return true;
}


static float4 dirFromSky(0.0f, -1.0f, 0.0f, 0.0f);

/// light directions (view space)
static float4 dirFromLight(0.577f, -0.577f, 0.577f, 0.0f);

/// light intensity
static float4 I_a(0.8f, 0.8f, 0.8f, 1.0f);        /// ambient
static float4 I_d(0.9f, 0.9f, 0.9f, 1.0f);        /// diffuse
static float4 I_s(0.9f, 0.9f, 0.9f, 1.0f);        /// specular
static float4 I_b(0.1f, 0.0f, 0.0f, 1.0f);        /// ground
static float4 I_c(0.9f, 0.9f, 1.0f, 1.0f);        /// sky

static float I_n = 32.0f;                         /// power

void MeshM2::draw()
{
    //
	graphicsModule()->setRenderState(RS_CULLMODE, CULL_NONE);

	//
    m_indexbuffer->bind();
    m_vertexbuffer->bind();

    //
	Batch * batch = m_batch;
	while (batch->shader)
    {
        IShader * shader = batch->shader;
        shader->begin();

		shader->setMatrix("projection", graphicsModule()->getProjectionMatrix());
		shader->setMatrix("view", graphicsModule()->getViewMatrix());

     	float4x4 transform;
		transform.setIdentity();
		
		shader->setMatrix("world", transform);

		shader->setVector("k_a", batch->ambient);
		shader->setVector("k_d", batch->diffuse);
		shader->setVector("k_s", batch->specular);

		shader->setVector("I_a", I_a);
		shader->setVector("I_d", I_d);
		shader->setVector("I_s", I_s);
		shader->setVector("I_b", I_b);
		shader->setVector("I_c", I_c);

		shader->setFloat("n", I_n);

		shader->setVector("dirFromSky", dirFromSky);
		shader->setVector("dirFromLight", dirFromLight);

        //
        switch (batch->flag)
        {
        case BM_OPAQUE:
            break;

        case BM_TRANSPARENT: // 1
            graphicsModule()->setRenderState(RS_ALPHATESTENABLE, TRUE);
            graphicsModule()->setRenderState(RS_ALPHAREF, 0xb2);
            graphicsModule()->setRenderState(RS_ALPHAFUNC, CMP_GREATEREQUAL);
            break;

        case BM_ALPHA_BLEND: // 2
            graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, TRUE);
            graphicsModule()->setRenderState(RS_SRCBLEND, BLEND_SRCALPHA);
            graphicsModule()->setRenderState(RS_DESTBLEND, BLEND_INVSRCALPHA);
            break;

        case BM_ADDITIVE: // 3
            graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, TRUE);
            graphicsModule()->setRenderState(RS_SRCBLEND, BLEND_SRCCOLOR);
            graphicsModule()->setRenderState(RS_DESTBLEND, BLEND_ONE);
            break;

        case BM_ADDITIVE_ALPHA: // 4
            graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, TRUE);
            graphicsModule()->setRenderState(RS_SRCBLEND, BLEND_SRCALPHA);
            graphicsModule()->setRenderState(RS_DESTBLEND, BLEND_ONE);
            break;

        case BM_MODULATE:	// 5
            graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, TRUE);
            graphicsModule()->setRenderState(RS_SRCBLEND, BLEND_DESTCOLOR);
            graphicsModule()->setRenderState(RS_DESTBLEND, BLEND_SRCCOLOR);
            break;
        case BM_MODULATEX2:	// 6, not sure if this is right
            graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, TRUE);
            graphicsModule()->setRenderState(RS_SRCBLEND, BLEND_DESTCOLOR);
            graphicsModule()->setRenderState(RS_DESTBLEND, BLEND_SRCCOLOR);
            break;

        default:
            graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, TRUE);
            graphicsModule()->setRenderState(RS_SRCBLEND, BLEND_SRCALPHA);
            graphicsModule()->setRenderState(RS_DESTBLEND, BLEND_INVSRCALPHA);
            break;
        }

        //
        for (int n = 0; n < MAX_SAMPLER_STAGE; n ++)
        {
            if (batch->sampler[n])
            {
                shader->setSampler(n, batch->sampler[n]);

                shader->setSamplerState(n, SAMP_MINFILTER, TEXF_ANISOTROPIC);
                shader->setSamplerState(n, SAMP_MAGFILTER, TEXF_ANISOTROPIC);
                shader->setSamplerState(n, SAMP_MIPFILTER, TEXF_LINEAR);
            }
        }

		//
        m_indexbuffer->draw(PT_TRIANGLELIST, batch->startIndex, batch->indexCount, m_vertexbuffer->getLength());

        shader->end();

        //
		graphicsModule()->setRenderState(RS_ALPHATESTENABLE, FALSE);
		graphicsModule()->setRenderState(RS_ALPHABLENDENABLE, FALSE);

        batch ++;
    }

    graphicsModule()->flush();
}

} //hydra
