/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#ifndef IGRAPHICS_H_
#define IGRAPHICS_H_

#include "kernel.h"

#include "vector.h"

#if defined(_XBOX)
#undef _NVCG
#endif //_XBOX

#if defined(_WIN32) || defined(__APPLE__) || defined(_LINUX)
	#if !defined(_NVCG)
	#define _NVCG
	#endif //_NVCG
#endif //_WIN32

#if defined(_NVCG)
	#include <Cg/cg.h>

	#if defined (_MSC_VER)
	#pragma comment(lib, "cg.lib")
	#endif //_MSC_VER
#endif //_NVCG

//
enum {
	FMT_UNKNOWN = 0,

	FMT_R8G8B8,
	FMT_A8R8G8B8,

	FMT_R5G6B5,
	FMT_A1R5G5B5,
	FMT_A4R4G4B4,

	FMT_A8,
	FMT_L8,
	FMT_A8L8,

	FMT_DXT1,
	FMT_DXT3,
	FMT_DXT5,

	FMT_INDEX16 = 0,				// 16-bit index buffer bit depth.
	FMT_INDEX32,					// 32-bit index buffer bit depth.
	FMT_FORCE_DWORD = 0x7fffffff,	// force 32-bit size enum
};

#define MAX_SAMPLER_STAGE	(8)

// The following defines the rendering states

// SHADEMODE
enum {
	SHADE_FLAT = 0,
	SHADE_GOURAUD,
	SHADE_PHONG,
	SHADE_FORCE_DWORD = 0x7fffffff, /* force 32-bit size enum */
};

// FILLMODE
enum {
	FILL_POINT = 0,
	FILL_WIREFRAME,
	FILL_SOLID,
	FILL_FORCE_DWORD = 0x7fffffff, /* force 32-bit size enum */
};

// BLEND
enum {
	BLEND_ZERO = 0,
	BLEND_ONE,
	BLEND_SRCCOLOR,
	BLEND_INVSRCCOLOR,
	BLEND_SRCALPHA,
	BLEND_INVSRCALPHA,
	BLEND_DESTALPHA,
	BLEND_INVDESTALPHA,
	BLEND_DESTCOLOR,
	BLEND_INVDESTCOLOR,
	BLEND_SRCALPHASAT,
	BLEND_FORCE_DWORD = 0x7fffffff, /* force 32-bit size enum */
};

// CULL
enum {
	CULL_NONE = 0,
	CULL_CW,
	CULL_CCW,
	CULL_FORCE_DWORD = 0x7fffffff, // force 32-bit size enum
};

// CMPFUNC
enum {
	CMP_NEVER = 0,
	CMP_LESS,
	CMP_EQUAL,
	CMP_LESSEQUAL,
	CMP_GREATER,
	CMP_NOTEQUAL,
	CMP_GREATEREQUAL,
	CMP_ALWAYS,
	CMP_FORCE_DWORD = 0x7fffffff, /* force 32-bit size enum */
};

// RENDERSTATETYPE
enum {
	RS_ZENABLE = 0,			// ZBUFFERTYPE (or TRUE/FALSE for legacy)
	RS_STENCILENABLE,		// BOOL enable/disable stenciling
	RS_ALPHATESTENABLE,		// TRUE to enable alpha tests
	RS_ALPHABLENDENABLE,	// TRUE to enable alpha blending
	RS_SRCBLEND,			// BLEND
	RS_DESTBLEND,			// BLEND
	RS_CULLMODE,			// CULL
	RS_SHADEMODE,			// SHADEMODE
	RS_FILLMODE,			// FILLMODE
	RS_ZFUNC,				// CMPFUNC
	RS_ALPHAFUNC,			// CMPFUNC
	RS_ALPHAREF,			// FIXED
	RS_FORCE_DWORD = 0x7fffffff, /* force 32-bit size enum */
};

// Values for SAMP_***FILTER texture stage states

// TEXTUREFILTERTYPE
enum {
	TEXF_NONE = 0,				// filtering disabled (valid for mip filter only)
	TEXF_POINT,					// nearest
	TEXF_LINEAR,				// linear interpolation
	TEXF_ANISOTROPIC,			// anisotropic
	TEXF_FORCE_DWORD = 0x7fffffff,   // force 32-bit size enum
};

// TEXTUREADDRESS
enum {
	TADDRESS_WRAP = 0,
	TADDRESS_CLAMP,
	TADDRESS_FORCE_DWORD     = 0x7fffffff, /* force 32-bit size enum */
};

// SAMPLERSTATETYPE
enum {
	SAMP_ADDRESSU = 0,		// TEXTUREADDRESS for U coordinate
	SAMP_ADDRESSV,			// TEXTUREADDRESS for V coordinate
	SAMP_MAGFILTER,			// TEXTUREFILTER filter to use for magnification
	SAMP_MINFILTER,			// TEXTUREFILTER filter to use for minification
	SAMP_MIPFILTER,			// TEXTUREFILTER filter to use between mipmaps during minification
	SAMP_FORCE_DWORD = 0x7fffffff,   // force 32-bit size enum
};

//
#define FONT_GLYPH_WIDTH		8
#define FONT_GLYPH_HEIGHT		9

namespace hydra
{

class IVertexbuffer;
class ISurface;
class ISampler;
class IShader;

class LIBAPI IGraphics
{
	DECLARE_DLL_MODULE(IGraphics)

public:
	IGraphics();
	virtual ~IGraphics();

public:
    bool open(HWND hwnd, int width, int height, int bpp, int samples, bool fullscreen);

	///
    virtual bool open(HWND hwnd);

	//
    virtual bool beginScene();
    virtual void endScene();
    virtual void present();

    virtual bool reset();

    //
    virtual void flush();

    //
	virtual void setViewport(int left, int top, int right, int bottom);

	//
	virtual void setRenderState(int state, unsigned long value) = 0;

	// TODO:
	virtual void setRenderTarget(unsigned long index, ISurface * surface) = 0;

	/// make_rhw (x,y, 0, 0)
	inline float3 make_rhw(int x, int y)
	{
		return float3(((float) x - m_hwidth) / m_hwidth, ((float) y - m_hheight) / -m_hheight, 0.0f);
	}
	
	inline float make_rhw_x(int x)
	{
		return (((float) x - m_hwidth) / m_hwidth);
	}
	inline float make_rhw_y(int y)
	{
		return (((float) y - m_hheight) / -m_hheight);
	}

	// text
public:
	/// text out
	void drawText(const fixed2 & pos, const char * text, unsigned long rgba = 0xffffffff);

	// property
public:
	//
	inline int getGlyphWidth()	{ return FONT_GLYPH_WIDTH; }
	inline int getGlyphHeight()	{ return FONT_GLYPH_HEIGHT; }

protected:
	bool createFont();

	// sprite
public:
	/// sprite
	void begin();
	void end();

	void draw(ISampler * sampler, const fixed2 & pos, unsigned long color = 0xffffffff);
	void draw(ISampler * sampler, const fixed2 & pos, const fixed2 & size, unsigned long color = 0xffffffff);
	void draw(ISampler * sampler, const fixed2 & pos, const fixed4 & rect, unsigned long color = 0xffffffff);

protected:
	/// font & sprite
	IVertexbuffer * m_vertexbuffer;
	int m_numVerts;

	ISampler * m_sampler;
	IShader * m_shader;

	// property
public:

	//
	void setViewMatrix(float4x4 & view)	{ m_view = view; }
	const float4x4 & getViewMatrix()			{ return m_view; }

	void setProjectionMatrix(float4x4 & projection)	{ m_projection = projection; }
	const float4x4 & getProjectionMatrix()				{ return m_projection; }

	//
	inline int getWidth() 	{ return m_width; }
	inline int getHeight() 	{ return m_height; }

protected:
	virtual void initialize();

#if defined(_NVCG)
public:
	/// nVidia Cg shader
	inline CGcontext getCgContext()	{ return m_cgContext; }

protected:
	/// nVidia Cg shader
	CGcontext m_cgContext;
#endif //_NVCG

protected:
	int m_width, m_height;
	int m_bpp, m_samples;
	bool m_fullscreen;

	float4x4 m_view;
	float4x4 m_projection;
	
	// half
	float m_hwidth, m_hheight;
};

DEFINE_DLL_MODULE(IGraphics, graphicsModule)

} //hydra

#endif //IGRAPHICS_H_
