/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "application.h"

#include "audio/iaudio.h"
#include "graphics/igraphics.h"
#include "system/iinput.h"
#include "system/iwindow.h"

#include "iview.h"

namespace hydra
{

#if defined(_WIN32)
	static void openProfile(const char * pathName);
	static void setSection(const char * sectionName);
	static DWORD getPrivateProfileInt(LPCTSTR lpKeyName, INT nDefault);
	static DWORD getPrivateProfileString(LPCTSTR lpKeyName, LPCTSTR lpDefault, LPTSTR lpReturnedString, DWORD dwSize);
#endif

//
Application::Application()
{
	m_input = NULL;
	m_window = NULL;
	m_view = NULL;

	//
	m_framesPerSecond = 0.0f;
	m_frameCounter = 0;
}
Application::~Application()
{
	//
	SAFE_RELEASE(m_input);
	SAFE_RELEASE(m_window);
}

bool Application::create()
{
	//
	if (!loadModule())
		return false;
		
	return true;
}

//
bool Application::loadModule()
{
	loadKernelModule("system");
	
	m_window = createWindow();
	if (!m_window->create(1, 1, "hydra"))
		return false;

	HWND hwnd = m_window->getHandle();

#if defined(_WIN32)
	char fullname[MAX_PATH];
	char module[MAX_PATH];
	openProfile("config.ini");

    // load audio module
	setSection("AUDIO");
	getPrivateProfileString("MODULE", "fmod4", module, MAX_PATH);

	strcpy(fullname, "audio.");
	strcat(fullname, module);
	
	loadKernelModule(fullname);
	if (!audioModule()->open(hwnd))
		return false;

	// load video module ("opengl")
	setSection("GRAPHICS");
	int width = getPrivateProfileInt("WIDTH", 1024);
	int height = getPrivateProfileInt("HEIGHT", 768);
	int bpp = getPrivateProfileInt("BPP", 32);
	int samples = getPrivateProfileInt("SAMPLES", 4);
	int fullscreen = getPrivateProfileInt("FULLSCREEN", 0);
	getPrivateProfileString("MODULE", "direct3d9", module, MAX_PATH);

	strcpy(fullname, "graphics.");
	strcat(fullname, module);
	
	loadKernelModule(fullname);
	if (!graphicsModule()->open(hwnd, width, height, bpp, samples, fullscreen ? true : false))
		return false;

    // load input module
	setSection("INPUT");
	getPrivateProfileString("MODULE", "input", module, MAX_PATH);

	strcpy(fullname, "input.");
	strcat(fullname, module);

	m_input = createInput(fullname);
	if (!m_input->open(hwnd))
		return false;

#else
	loadKernelModule("audio.fmod4");
	if (!audioModule()->open(hwnd))
		return false;

	int width = 1024;
	int height = 768;
	int bpp = 32;
	int fullscreen = 0;

	loadKernelModule("graphics.opengl");
	if (!graphicsModule()->open(hwnd, width, height, bpp, 4, fullscreen ? true : false))
		return false;

	m_input = createInput();
	if (!m_input->open(hwnd))
		return false;
#endif

	return true;
}

//
bool Application::update()
{
	// update frame per second
	float time, deltaTime;
	static float lastFrameTime, lastAnimTime;

	static bool firstCall = true;
	if (firstCall)
	{
		lastFrameTime = lastAnimTime = TIMER;
		firstCall = false;
	}

	time = TIMER;
	if (time > (lastFrameTime + 1000))
	{
		m_framesPerSecond = (float) m_frameCounter;
		m_frameCounter = 0;

		lastFrameTime = time;
	}

	deltaTime = (time - lastAnimTime) * 0.001f;
	lastAnimTime = time;

	//
	audioModule()->update();
	
	m_input->update();
	
	return onUpdate(deltaTime);
}
void Application::draw()
{
	if (graphicsModule()->beginScene())
	{
		onDraw();

		graphicsModule()->endScene();
	}
}
void Application::present()
{
	graphicsModule()->present();

	//
	m_frameCounter ++;
}

int Application::run()
{
	return m_window->run();
}

//
bool Application::onUpdate(float deltaTime)
{
	return m_view->update(deltaTime);
}

//
void Application::onDraw()
{
	m_view->draw();
}

#if defined(_WIN32)
//
static char m_pathName[MAX_PATH];
static char m_sectionName[MAX_PATH];
static void openProfile(const char * pathName)
{
	::GetCurrentDirectory(MAX_PATH, m_pathName);

	::strcat(m_pathName, "\\");
	::strcat(m_pathName, pathName);
}
static void setSection(const char * sectionName)
{
	::strcpy(m_sectionName, sectionName);
}

//
static DWORD getPrivateProfileInt(LPCTSTR lpKeyName, INT nDefault)
{
	return ::GetPrivateProfileInt(m_sectionName, 
		lpKeyName, 
		nDefault, 
		m_pathName);
}
static DWORD getPrivateProfileString(LPCTSTR lpKeyName, LPCTSTR lpDefault, LPTSTR lpReturnedString, DWORD dwSize)
{
	return ::GetPrivateProfileString(m_sectionName, 
		lpKeyName,
		lpDefault, 
		lpReturnedString, 
		dwSize,
		m_pathName);
}
#endif //_WIN32

} //hydra
