/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "graphics/igraphics.h"

#include "graphics/ivertexbuffer.h"
#include "graphics/isampler.h"
#include "graphics/ishader.h"

#include "graphics/imageDDS.h"

#include "kernel.h"

namespace hydra
{

IMPLEMENT_DLL_MODULE(IGraphics)

IGraphics::IGraphics()
{
	//
	pmodule = this;

	//
	m_vertexbuffer = NULL;
	m_sampler = NULL;
	m_shader = NULL;

	//
	m_bpp = 32;
	m_fullscreen = false;

#if defined(_NVCG)
	m_cgContext = NULL;
#endif //_NVCG
}
IGraphics::~IGraphics()
{
	//TRACE("# IGraphics::~IGraphics()")

	// register image modules
	unregisterKernelModule(ImageDDS);

#if defined(_NVCG)
	// Since we are recreating all our programs based on the device type,
	// we destroy our programs every time we destroy the device
	if (m_cgContext)
	{
		cgDestroyContext(m_cgContext);
		m_cgContext = NULL;
	}
#endif //_NVCG

	//
	pmodule = NULL;
}

//
bool IGraphics::open(HWND hwnd)
{
	//TRACE("# IGraphics::open()")

#if defined(_NVCG)
	// Set the device and pick an appropriate profile
	m_cgContext = cgCreateContext();
	if (!m_cgContext)
		return false;

	TRACE(">> nVidia Cg Version: %d.%d\n", CG_VERSION_NUM/1000, CG_VERSION_NUM-1000)

#if defined(_WIN32)
	cgSetParameterSettingMode(m_cgContext, CG_DEFERRED_PARAMETER_SETTING);
#endif //_WIN32
#endif //_NVCG

    //
    setViewport(0, 0, m_width, m_height);

    //
    initialize();

	// register image modules
	registerKernelModule(ImageDDS);

    //
    if (!createFont())
        return false;

	return true;
}

bool IGraphics::open(HWND hwnd, int width, int height, int bpp, int samples, bool fullscreen)
{
	m_width = width;
	m_height = height;

	m_bpp = bpp;
	m_samples = samples;

	m_fullscreen = fullscreen;

	return open(hwnd);
}

//
bool IGraphics::beginScene()
{
	return true;
}
void IGraphics::endScene()
{

}
void IGraphics::present()
{

}

// for direct3d9
bool IGraphics::reset()
{
	return true;
}

// for ps3
void IGraphics::flush()
{

}

void IGraphics::setViewport(int left, int top, int right, int bottom)
{
	m_width = right;
	m_height = bottom;

	//
	m_hwidth = (float) m_width * 0.5f;
	m_hheight = (float) m_height * 0.5f;
}

//
void IGraphics::initialize()
{

}

//
#define FONT_TEXTURE_WIDTH		128
#define FONT_TEXTURE_HEIGHT		128
#define FONT_DATA_SIZE			(FONT_TEXTURE_WIDTH * FONT_TEXTURE_HEIGHT)
#define FONT_TAB_SIZE			4

static unsigned char m_fontData[FONT_DATA_SIZE] = {
#include "fontdata.inl"
};

bool IGraphics::createFont()
{
	//TRACE("# IGraphics::createFont()")

	//
	unsigned long fvf = FVF_XYZ | FVF_DIFFUSE | FVF_TEX1;
	m_vertexbuffer = createVertexbuffer(MAX_PATH * 6, fvf);
	if (!m_vertexbuffer)
		return false;

    //create shader
#if defined(_PS3) && !defined(_USEPSGL)
	char * vptr = (char *) &_binary_font_vp_vpo_start;
	char * fptr = (char *) &_binary_font_fp_fpo_start;

	m_shader = createShader("font.cg");
	if (!m_shader->loadFromMemory(vptr, fptr))
		return false;
//#endif //_PS3
#else
	m_shader = loadShader("shaders/font.cg");
	if (!m_shader)
		return false;
#endif //else

	//create sampler
#if defined(_PS3) && !defined(_USEPSGL)
	m_sampler = createSampler(FONT_TEXTURE_WIDTH, FONT_TEXTURE_HEIGHT, FMT_L8, 0);
	if (!m_sampler)
		return false;
//#endif //_PS3
#else
	m_sampler = createSampler(FONT_TEXTURE_WIDTH, FONT_TEXTURE_HEIGHT, FMT_A8L8, 0);
	if (!m_sampler)
		return false;
#endif //else

	unsigned char * bits = (unsigned char *) m_sampler->map();
	unsigned char * fontData = (unsigned char *) m_fontData;

	for (int n=0; n<FONT_DATA_SIZE; n++)
	{
#if defined(_PS3) && !defined(_USEPSGL)
		*bits ++ = *fontData;
//#endif //_PS3
#elif defined(_XBOX)
		*bits ++ = 0xff;
		*bits ++ = *fontData;
//#endif //_XBOX
#else
		*bits ++ = *fontData;
		*bits ++ = 0xff;
#endif //else

		fontData ++;
	}
	m_sampler->unmap();

	return true;
}

//
static inline float make_left(char c)
{
	return (float)(((unsigned char) c % 16) * FONT_GLYPH_WIDTH) / (float) FONT_TEXTURE_WIDTH;
}
static inline float make_right(char c)
{
	return (float)((((unsigned char) c % 16) * FONT_GLYPH_WIDTH) + FONT_GLYPH_WIDTH) / (float) FONT_TEXTURE_WIDTH;
}
static inline float make_top(char c)
{
	return (float)(((((unsigned char) c / 16) - 2) * FONT_GLYPH_HEIGHT) + 1) / (float) FONT_TEXTURE_HEIGHT;
}
static inline float make_bottom(char c)
{
	return (float)(((((unsigned char) c / 16) - 2) * FONT_GLYPH_HEIGHT) + FONT_GLYPH_HEIGHT + 1) / (float) FONT_TEXTURE_HEIGHT;
}

//
struct VERTEXDATA {
	float3 position;		// Homogeneous coordinates
	unsigned long color;
	float2 tex0;			// Texture coordinates
};


void IGraphics::drawText(const fixed2 & pos, const char * text, unsigned long rgba/*= 0xffffffff*/)
{
	int x = pos.x;
	int y = pos.y;

	m_numVerts = 0;

	VERTEXDATA * verts = (VERTEXDATA *) m_vertexbuffer->map();

	char * ptr = (char *) text;
	while (*ptr)
	{
		if (*ptr == '\n')
		{
			x = pos.x;
			y += FONT_GLYPH_HEIGHT + 1;
		}
		else if (*ptr == '\t')
		{
			x += FONT_TAB_SIZE * (FONT_GLYPH_WIDTH + 1);
		}
		else
		{
			float rx = make_rhw_x(x);
			float ry = make_rhw_y(y);
			float rw = make_rhw_x(x + FONT_GLYPH_WIDTH);
			float rh = make_rhw_y(y + FONT_GLYPH_HEIGHT);
			
			float tx = make_left(*ptr);
			float ty = make_top(*ptr);
			float tw = make_right(*ptr);
			float th = make_bottom(*ptr);
			
			// left, top
			verts->position = float3(rx, ry, 0);
			verts->color = rgba;
			verts->tex0 = float2(tx, ty);
			verts ++;

			// right, top
			verts->position = float3(rw, ry, 0);
			verts->color = rgba;
			verts->tex0 = float2(tw, ty);
			verts ++;

			// right, bottom
			verts->position = float3(rw, rh, 0);
			verts->color = rgba;
			verts->tex0 = float2(tw, th);
			verts ++;

			// right, bottom
			verts->position = float3(rw, rh, 0);
			verts->color = rgba;
			verts->tex0 = float2(tw, th);
			verts ++;

			// left, bottom
			verts->position = float3(rx, rh, 0);
			verts->color = rgba;
			verts->tex0 = float2(tx, th);
			verts ++;

			// left, top
			verts->position = float3(rx, ry, 0);
			verts->color = rgba;
			verts->tex0 = float2(tx, ty);
			verts ++;

			m_numVerts += 6;

			x += FONT_GLYPH_WIDTH + 1;
		}

		//
		if ((x + (FONT_GLYPH_WIDTH + 1)) >= m_width)
		{
			x = pos.x;
			y += FONT_GLYPH_HEIGHT + 1;
		}

		ptr ++;
	}
	m_vertexbuffer->unmap();

	//
	setRenderState(RS_ZENABLE, FALSE);
	setRenderState(RS_CULLMODE, CULL_NONE);
	setRenderState(RS_STENCILENABLE, FALSE);

	setRenderState(RS_ALPHABLENDENABLE, TRUE);
	setRenderState(RS_SRCBLEND, BLEND_SRCCOLOR);
	setRenderState(RS_DESTBLEND, BLEND_ONE);

	//
	m_shader->begin();
	m_shader->setSampler(0, m_sampler);

    //
	m_shader->setSamplerState(0, SAMP_MINFILTER, TEXF_POINT);
 	m_shader->setSamplerState(0, SAMP_MAGFILTER, TEXF_POINT);
 	m_shader->setSamplerState(0, SAMP_MIPFILTER, TEXF_NONE);

	//
	m_vertexbuffer->bind();
	m_vertexbuffer->draw(PT_TRIANGLELIST, 0, m_numVerts);

	m_shader->end();

	flush();


	//
	setRenderState(RS_ZENABLE, TRUE);
	setRenderState(RS_CULLMODE, CULL_CW);
	setRenderState(RS_STENCILENABLE, TRUE);

	setRenderState(RS_ALPHABLENDENABLE, FALSE);
	setRenderState(RS_ALPHATESTENABLE, FALSE);

	// debug
	/*begin();

	float4x4 transform;
	transform.setIdentity();
	
	draw(m_sampler, fixed2(100, 200));
	end();*/
}

// sprite
void IGraphics::begin()
{
	setRenderState(RS_ZENABLE, FALSE);
	setRenderState(RS_CULLMODE, CULL_NONE);
	setRenderState(RS_STENCILENABLE, FALSE);

	//setRenderState(RS_ALPHABLENDENABLE, TRUE);
	//setRenderState(RS_SRCBLEND, BLEND_SRCCOLOR);
	//setRenderState(RS_DESTBLEND, BLEND_ONE);

	m_shader->begin();
}
void IGraphics::end()
{
	m_shader->end();

	flush();

	//
	setRenderState(RS_ZENABLE, TRUE);
	setRenderState(RS_CULLMODE, CULL_CW);
	setRenderState(RS_STENCILENABLE, TRUE);

	setRenderState(RS_ALPHABLENDENABLE, FALSE);
	setRenderState(RS_ALPHATESTENABLE, FALSE);
}

void IGraphics::draw(ISampler * sampler, const fixed2 & pos, unsigned long color /* = 0xffffffff */)
{
	// (left, right)
	// 0------1
	// |      |
	// 3------2 (right, bottom)
	float rx = make_rhw_x(pos.x);
	float ry = make_rhw_y(pos.y);
	float rw = make_rhw_x(pos.x + sampler->getWidth());
	float rh = make_rhw_y(pos.y + sampler->getHeight());
	
	VERTEXDATA * verts = (VERTEXDATA *) m_vertexbuffer->map();
	verts[0].position = float3(rx, ry, 0);
	verts[0].color = color;
	verts[0].tex0 = float2(0, 0);

	verts[1].position = float3(rw, ry, 0);
	verts[1].color = color;
	verts[1].tex0 = float2(1, 0);

	verts[2].position = float3(rw, rh, 0);
	verts[2].color = color;
	verts[2].tex0 = float2(1, 1);

	verts[3].position = float3(rx, rh, 0);
	verts[3].color = color;
	verts[3].tex0 = float2(0, 1);
	m_vertexbuffer->unmap();

    //
	m_shader->setSampler(0, sampler);
	
	//
	m_shader->setSamplerState(0, SAMP_MAGFILTER, TEXF_POINT);
	m_shader->setSamplerState(0, SAMP_MINFILTER, TEXF_POINT);
	m_shader->setSamplerState(0, SAMP_MIPFILTER, TEXF_NONE);

	m_vertexbuffer->bind();
	m_vertexbuffer->draw(PT_TRIANGLEFAN, 0, 4);
}

void IGraphics::draw(ISampler * sampler, const fixed2 & pos, const fixed2 & size, unsigned long color)
{
	// (left, right)
	// 0------1
	// |      |
	// 3------2 (right, bottom)
	VERTEXDATA * verts = (VERTEXDATA *) m_vertexbuffer->map();
	verts[0].position = make_rhw(pos.x, pos.y);
	verts[0].color = color;
	verts[0].tex0 = float2(0, 0);

	verts[1].position = make_rhw(pos.x + size.x, pos.y);
	verts[1].color = color;
	verts[1].tex0 = float2(1, 0);

	verts[2].position = make_rhw(pos.x + size.x, pos.y + size.y);
	verts[2].color = color;
	verts[2].tex0 = float2(1, 1);

	verts[3].position = make_rhw(pos.x, pos.y + size.y);
	verts[3].color = color;
	verts[3].tex0 = float2(0, 1);
	m_vertexbuffer->unmap();

	//
	m_shader->setSampler(0, sampler);

	//
	m_shader->setSamplerState(0, SAMP_MAGFILTER, TEXF_POINT);
	m_shader->setSamplerState(0, SAMP_MINFILTER, TEXF_POINT);
	m_shader->setSamplerState(0, SAMP_MIPFILTER, TEXF_NONE);

	m_vertexbuffer->bind();
	m_vertexbuffer->draw(PT_TRIANGLEFAN, 0, 4);
}

void IGraphics::draw(ISampler * sampler, const fixed2 & pos, const fixed4 & rect, unsigned long color)
{
	// (left, right)
	// 0------1
	// |      |
	// 3------2 (right, bottom)
	float right = (float) sampler->getWidth();
	float bottom = (float) sampler->getHeight();

	VERTEXDATA * verts = (VERTEXDATA *) m_vertexbuffer->map();
	verts[0].position = make_rhw(pos.x, pos.y);
	verts[0].color = color;
	verts[0].tex0 = float2(rect.left / right, rect.top / bottom);

	verts[1].position = make_rhw(pos.x + rect.right, pos.y);
	verts[1].color = color;
	verts[1].tex0 = float2((rect.left + rect.right) / right, rect.top / bottom);

	verts[2].position = make_rhw(pos.x + rect.right, pos.y + rect.bottom);
	verts[2].color = color;
	verts[2].tex0 = float2((rect.left + rect.right) / right, (rect.top + rect.bottom) / bottom);

	verts[3].position = make_rhw(pos.x, pos.y + rect.bottom);
	verts[3].color = color;
	verts[3].tex0 = float2(rect.left / right, (rect.top + rect.bottom) / bottom);
	m_vertexbuffer->unmap();

	//
	m_shader->setSampler(0, sampler);

	//
	m_shader->setSamplerState(0, SAMP_MAGFILTER, TEXF_POINT);
	m_shader->setSamplerState(0, SAMP_MINFILTER, TEXF_POINT);
	m_shader->setSamplerState(0, SAMP_MIPFILTER, TEXF_NONE);

	m_vertexbuffer->bind();
	m_vertexbuffer->draw(PT_TRIANGLEFAN, 0, 4);
}

} //hydra
