/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "graphics/isampler.h"

#include "graphics/iimage.h"
#include "graphics/igraphics.h"

#include "kernel.h"

namespace hydra
{
IMPLEMENT_BASE_MODULE(ISampler, usr/lib/graphics/sampler)

ISampler::ISampler()
{

}

static const char * __formats[] =
{
	"FMT_UNKNOWN",

	"FMT_R8G8B8",
	"FMT_A8R8G8B8",

	"FMT_R5G6B5",
	"FMT_A1R5G5B5",
	"FMT_A4R4G4B4",

	"FMT_A8",
	"FMT_L8",
	"FMT_A8L8",

	"FMT_DXT1",
	"FMT_DXT3",
	"FMT_DXT5",
};
bool ISampler::create(int width, int height, unsigned long format, int mipLevels)
{
	//TRACE("# ISampler::create(%d, %d, %s, %d)\n", width, height, __formats[format], mipLevels)

	int bitsPerPixels = 1;
	switch (format)
	{
	case FMT_R8G8B8:
		bitsPerPixels = 3;
		break;
	case FMT_A8R8G8B8:
		bitsPerPixels = 4;
		break;

	case FMT_R5G6B5:
	case FMT_A1R5G5B5:
	case FMT_A4R4G4B4:
		bitsPerPixels = 2;
		break;

	case FMT_A8:
		bitsPerPixels = 1;
		break;

	case FMT_L8:
		bitsPerPixels = 1;
		break;

	case FMT_A8L8:
		bitsPerPixels = 2;
		break;

	case FMT_DXT1:
		bitsPerPixels = 8;
		break;

	case FMT_DXT3:
		bitsPerPixels = 16;
		break;

	case FMT_DXT5:
		bitsPerPixels = 16;
		break;
	}

	m_format = format;

	m_width = width;
	m_height = height;

	m_pitch = width * bitsPerPixels;

	//
	m_mipLevels = mipLevels;

	//
	m_linearSize = 0;

	//
	int w = width;
	int h = height;
	int levels = m_mipLevels ? m_mipLevels : 1;

	for (int n=0; n<levels; n++)
	{
		int linearSize;
		if (format > FMT_A8L8)
			linearSize = (((w + 3) >> 2) * ((h + 3) >> 2)) * bitsPerPixels;
		else
			linearSize = (h + 1) * (w * bitsPerPixels);

		m_linearSize += linearSize;

		w = max(w >> 1, 1);
		h = max(h >> 1, 1);
	}

	//TRACE(">> linerSize = %d\n", m_linearSize)

	return true;
}

//
bool ISampler::loadFromFile(const char * pathName)
{
	//TRACE("# ISampler::loadFromFile(%s)\n", pathName)

	IImage * image = loadImage(pathName);
	if (!image)
		return false;

	decode(image);
	SAFE_RELEASE(image);

	return true;
}

//
bool ISampler::decode(IImage * image)
{
	if (!create(image->getWidth(), image->getHeight(), image->getFormat(), image->getMipLevels()))
		return false;

	//
	char * bits = (char *) map();
	char * src = (char *) image->getBits();

	switch (image->getFormat())
	{
	case FMT_R8G8B8:
	case FMT_A8R8G8B8:
		for (int i = 0; i < image->getHeight(); i ++)
		{
			memcpy(bits, src, image->getPitch());

			//
			bits += getLockPitch();
			src += image->getPitch();
		}
		break;

	case FMT_DXT1:
	case FMT_DXT3:
	case FMT_DXT5:
		memcpy(bits, src, image->getLinerSize());
		break;
	}
	unmap();

	return true;
/*
#if defined(_WIN32)
		LPDIRECT3DSURFACE9 surface;
		m_texture->GetSurfaceLevel(0, &surface);

		D3DSURFACE_DESC desc;
		surface->GetDesc(&desc);
		surface->Release();

		switch (desc.Format)
		{
		case D3DFMT_X8R8G8B8:
		case D3DFMT_A8R8G8B8:
			destSkipBytes = 4;
			break;
		default:
			destSkipBytes = 3;
		}
#endif //_WIN32
*/
	return true;
}

} //hydra
