/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "HID.h"

#if defined(__APPLE__)

static void HIDGetElementsCFArrayHandler(const void * value, void * parameter);

// returns current value for element, polling element
// will return 0 on error conditions which should be accounted for by application
SInt32 HIDGetElementValue(recDevice * device, recElement * element)
{
    IOReturn result;
    IOHIDEventStruct hidEvent;
    hidEvent.value = 0;

    if (device != NULL && element != NULL && device->interface != NULL)
	{
        result = (*(device->interface))->getElementValue(device->interface, element->cookie, &hidEvent);
        if (result == kIOReturnSuccess)
		{
            // record min and max for auto calibration
            if (hidEvent.value < element->minReport)
                element->minReport = hidEvent.value;
				
            if (hidEvent.value > element->maxReport)
                element->maxReport = hidEvent.value;
        }
    }

    // auto user scale
    return hidEvent.value;
}

SInt32 HIDScaledCalibratedValue(recDevice * device, recElement * element, long min, long max)
{
    float deviceScale = max - min;
    float readScale = element->maxReport - element->minReport;
    SInt32 value = HIDGetElementValue(device, element);
    if (readScale == 0)
        return value;           /* no scaling at all */
    else
        return ((value - element->minReport) * deviceScale / readScale) + min;
}

static void HIDRemovalCallback(void * target, IOReturn result, void * refcon, void * sender)
{
    recDevice * device = (recDevice *) refcon;
    device->removed = 1;
    device->uncentered = 1;
}

// create and open an interface to device, required prior to extracting values or building queues.
// Note: appliction now owns the device and must close and release it prior to exiting
static IOReturn HIDCreateOpenDeviceInterface(io_object_t hidDevice, recDevice * device)
{
	IOReturn result;
	
    if (device->interface == NULL)
	{
		SInt32 score;
		IOCFPlugInInterface ** plugInInterface;
        
		result = IOCreatePlugInInterfaceForService(hidDevice, kIOHIDDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
        if (kIOReturnSuccess == result)
		{
            // call a method of the intermediate plug-in to create the device interface
            HRESULT plugInResult = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOHIDDeviceInterfaceID), (void *) &device->interface);
            if (plugInResult != S_OK)
			{
                //TRACE0("Couldn’t query HID class device interface from plugInInterface\n")
			}
			
            (*plugInInterface)->Release(plugInInterface);
        }
		else
		{
            //TRACE0("Failed to create **plugInInterface via IOCreatePlugInInterfaceForService.\n")
		}
    }
	
    if (device->interface != NULL)
	{
        result = (*(device->interface))->open(device->interface, 0);
        if (result != kIOReturnSuccess)
		{
            //TRACE0("Failed to open pDevice->interface via open.\n")
		}
        else
            (*(device->interface))->setRemovalCallback(device->interface, HIDRemovalCallback, device, device);
    }

    return result;
}

// handles retrieval of element information from arrays of elements in device IO registry information
static void HIDGetElements(CFTypeRef refElementCurrent, recDevice * device)
{
    CFTypeID type = CFGetTypeID(refElementCurrent);
    if (type == CFArrayGetTypeID())
	{   
        CFRange range = {
			0,
			CFArrayGetCount(refElementCurrent)
		};
        
		// countElementsCFArrayHandler called for each array member
        CFArrayApplyFunction(refElementCurrent, range, HIDGetElementsCFArrayHandler, device);
    }
}

// handles extracting element information from element collection CF types
// used from top level element decoding and hierarchy deconstruction to flatten device element list
static void HIDGetCollectionElements(CFMutableDictionaryRef deviceProperties, recDevice * device)
{
    CFTypeRef refElementTop = CFDictionaryGetValue(deviceProperties, CFSTR(kIOHIDElementKey));
    if (refElementTop)
        HIDGetElements(refElementTop, device);
}

// extracts actual specific element information from each element CF dictionary entry
static void HIDGetElementInfo(CFTypeRef refElement, recElement * element)
{
    long number;
    CFTypeRef refType;

    refType = CFDictionaryGetValue(refElement, CFSTR(kIOHIDElementCookieKey));
    if (refType && CFNumberGetValue(refType, kCFNumberLongType, &number))
        element->cookie = (IOHIDElementCookie) number;
		
    refType = CFDictionaryGetValue(refElement, CFSTR(kIOHIDElementMinKey));
    if (refType && CFNumberGetValue(refType, kCFNumberLongType, &number))
        element->minReport = element->min = number;
		
    element->maxReport = element->min;
    refType = CFDictionaryGetValue(refElement, CFSTR(kIOHIDElementMaxKey));
    if (refType && CFNumberGetValue(refType, kCFNumberLongType, &number))
        element->maxReport = element->max = number;
}

// examines CF dictionary vlaue in device element hierarchy to determine if it is element of interest or a collection of more elements
// if element of interest allocate storage, add to list and retrieve element specific info
// if collection then pass on to deconstruction collection into additional individual elements
static void HIDAddElement(CFTypeRef refElement, recDevice * device)
{
    recElement * element = NULL;
    recElement ** headElement = NULL;
    long elementType, usagePage, usage;
    CFTypeRef refElementType = CFDictionaryGetValue(refElement, CFSTR(kIOHIDElementTypeKey));
    CFTypeRef refUsagePage = CFDictionaryGetValue(refElement, CFSTR(kIOHIDElementUsagePageKey));
    CFTypeRef refUsage = CFDictionaryGetValue(refElement, CFSTR(kIOHIDElementUsageKey));

    if ((refElementType) && (CFNumberGetValue(refElementType, kCFNumberLongType, &elementType)))
	{
        // look at types of interest
        if ((elementType == kIOHIDElementTypeInput_Misc) ||
            (elementType == kIOHIDElementTypeInput_Button) ||
            (elementType == kIOHIDElementTypeInput_Axis))
		{
            if (refUsagePage && CFNumberGetValue(refUsagePage, kCFNumberLongType, &usagePage) && refUsage &&
				CFNumberGetValue(refUsage, kCFNumberLongType, &usage))
			{
				// only interested in kHIDPage_GenericDesktop and kHIDPage_Button
                switch (usagePage)
				{    
                case kHIDPage_GenericDesktop:
                    {
						// look at usage to determine function
                        switch (usage)
						{        
                        case kHIDUsage_GD_X:
                        case kHIDUsage_GD_Y:
                        case kHIDUsage_GD_Z:
                        case kHIDUsage_GD_Rx:
                        case kHIDUsage_GD_Ry:
                        case kHIDUsage_GD_Rz:
                        case kHIDUsage_GD_Slider:
                        case kHIDUsage_GD_Dial:
                        case kHIDUsage_GD_Wheel:
                            element = (recElement *) NewPtrClear(sizeof(recElement));
                            if (element)
							{
                                device->axes++;
                                headElement = &(device->firstAxis);
                            }
                            break;
                        case kHIDUsage_GD_Hatswitch:
                            element = (recElement *) NewPtrClear(sizeof(recElement));
                            if (element)
							{
                                device->hats++;
                                headElement = &(device->firstHat);
                            }
                            break;
                        }
                    }
                    break;
                case kHIDPage_Button:
                    element = (recElement *) NewPtrClear(sizeof(recElement));
                    if (element)
					{
                        device->buttons++;
                        headElement = &(device->firstButton);
                    }
                    break;
                default:
                    break;
                }
            }
        }
		else if (kIOHIDElementTypeCollection == elementType)
		{
            HIDGetCollectionElements((CFMutableDictionaryRef) refElement, device);
		}
    }

	// add to list
    if (element && headElement)
	{       
        device->elements++;
        if (NULL == *headElement)
            *headElement = element;
        else
		{
            recElement * elementPrevious;
			recElement * elementCurrent = *headElement;
            while (elementCurrent)
			{
                elementPrevious = elementCurrent;
                elementCurrent = elementPrevious->next;
            }
            elementPrevious->next = element;
        }
        element->next = NULL;
        HIDGetElementInfo(refElement, element);
    }
}

// collects information from each array member in device element list (each array memeber = element) 
static void HIDGetElementsCFArrayHandler(const void * value, void * parameter)
{
    if (CFGetTypeID(value) == CFDictionaryGetTypeID())
        HIDAddElement(value, (recDevice *) parameter);
}

// use top level element usage page and usage to discern device usage page and usage setting appropriate vlaues in device record
static void HIDTopLevelElementHandler(const void * value, void * parameter)
{
    CFTypeRef refCF = 0;
    if (CFGetTypeID(value) != CFDictionaryGetTypeID())
        return;
		
    refCF = CFDictionaryGetValue(value, CFSTR(kIOHIDElementUsagePageKey));
    if (!CFNumberGetValue(refCF, kCFNumberLongType, &((recDevice *) parameter)->usagePage))
	{
        //TRACE0("CFNumberGetValue error retrieving pDevice->usagePage.\n")
	}
		
    refCF = CFDictionaryGetValue(value, CFSTR(kIOHIDElementUsageKey));
    if (!CFNumberGetValue(refCF, kCFNumberLongType, &((recDevice *) parameter)->usage))
	{
        //TRACE0("CFNumberGetValue error retrieving pDevice->usage.\n")
	}
}

// extracts device info from CF dictionary records in IO registry
static void HIDGetDeviceInfo(io_object_t hidDevice, CFMutableDictionaryRef hidProperties, recDevice * device)
{
    CFMutableDictionaryRef usbProperties = 0;
    io_registry_entry_t parent1, parent2;

    // Mac OS X currently is not mirroring all USB properties to HID page so need to look at USB device page also
    // get dictionary for usb properties: step up two levels and get CF dictionary for USB properties
    if ((IORegistryEntryGetParentEntry(hidDevice, kIOServicePlane, &parent1) == KERN_SUCCESS) &&
        (IORegistryEntryGetParentEntry(parent1, kIOServicePlane, &parent2) == KERN_SUCCESS) &&
        (IORegistryEntryCreateCFProperties(parent2, &usbProperties, kCFAllocatorDefault, kNilOptions) == KERN_SUCCESS))
	{
        if (usbProperties)
		{
            CFTypeRef refCF = 0;
            // get device info
            // try hid dictionary first, if fail then go to usb dictionary

            // get product name
            refCF = CFDictionaryGetValue(hidProperties, CFSTR(kIOHIDProductKey));
            if (!refCF)
                refCF = CFDictionaryGetValue(usbProperties, CFSTR("USB Product Name"));
            
			if (refCF)
			{
                if (!CFStringGetCString(refCF, device->product, 256, CFStringGetSystemEncoding()))
                {
				    //TRACE0("CFStringGetCString error retrieving pDevice->product.\n")
				}
            }

            // get usage page and usage
            refCF = CFDictionaryGetValue(hidProperties, CFSTR(kIOHIDPrimaryUsagePageKey));
            if (refCF)
			{
                if (!CFNumberGetValue(refCF, kCFNumberLongType, &device->usagePage))
				{
                    //TRACE0("CFNumberGetValue error retrieving pDevice->usagePage.\n")
				}
                
				refCF = CFDictionaryGetValue(hidProperties, CFSTR(kIOHIDPrimaryUsageKey));
                if (refCF)
				{
                    if (!CFNumberGetValue(refCF, kCFNumberLongType, &device->usage))
					{
                        //TRACE0("CFNumberGetValue error retrieving pDevice->usage.\n")
					}
				}
            }

			// get top level element HID usage page or usage
            if (refCF == NULL)
			{        
                // use top level element instead
                CFTypeRef refCFTopElement = 0;
                refCFTopElement = CFDictionaryGetValue(hidProperties, CFSTR(kIOHIDElementKey));

                // refCFTopElement points to an array of element dictionaries
                CFRange range = { 0, CFArrayGetCount(refCFTopElement) };
                CFArrayApplyFunction(refCFTopElement, range, HIDTopLevelElementHandler, device);
            }

            CFRelease(usbProperties);
        }
		else
		{
            //TRACE0("IORegistryEntryCreateCFProperties failed to create usbProperties.\n")
		}

        if (IOObjectRelease(parent2) != kIOReturnSuccess)
		{
            //TRACE0("IOObjectRelease error with parent2.\n")
		}
        if (IOObjectRelease(parent1) != kIOReturnSuccess)
		{
            //TRACE0("IOObjectRelease error with parent1.\n")
		}
    }
}

recDevice * HIDBuildDevice(io_object_t hidDevice)
{
    recDevice * device = (recDevice *) NewPtrClear(sizeof(recDevice));
    if (device)
	{
        // get dictionary for HID properties
        CFMutableDictionaryRef hidProperties = 0;
        kern_return_t result = IORegistryEntryCreateCFProperties(hidDevice, &hidProperties, kCFAllocatorDefault, kNilOptions);
        if ((result == KERN_SUCCESS) && hidProperties)
		{
            // create device interface
            result = HIDCreateOpenDeviceInterface(hidDevice, device);
            if (result == kIOReturnSuccess)
			{
				// hidDevice used to find parents in registry tree
                HIDGetDeviceInfo(hidDevice, hidProperties, device);
                HIDGetCollectionElements(hidProperties, device);
            }
			else
			{
                DisposePtr((Ptr) device);
                device = NULL;
            }
            CFRelease(hidProperties);
        }
		else
		{
            DisposePtr((Ptr) device);
            device = NULL;
        }
    }
	
    return device;
}

// closes and releases interface to device, should be done prior to exting application
// Note: will have no affect if device or interface do not exist
// application will "own" the device if interface is not closed
// (device may have to be plug and re-plugged in different location to get it working again without a restart)
static IOReturn HIDCloseReleaseInterface(recDevice * device)
{
    IOReturn result = kIOReturnSuccess;

    if ((device != NULL) && (device->interface != NULL))
	{
        // close the interface
        result = (*(device->interface))->close(device->interface);
        if (result == kIOReturnNotOpen)
		{
            // do nothing as device was not opened, thus can't be closed
        }
		else if (result != kIOReturnSuccess)
		{
            //TRACE0("Failed to close IOHIDDeviceInterface.\n")
		}
		
        // release the interface
        result = (*(device->interface))->Release(device->interface);
        if (result != kIOReturnSuccess)
		{
            //TRACE0("Failed to release IOHIDDeviceInterface.\n")
		}
        device->interface = NULL;
    }
	
    return result;
}

// disposes of the element list associated with a device and the memory associated with the list
static void HIDDisposeElementList(recElement ** elements)
{
    recElement * element = *elements;
    while (element)
	{
        recElement * next = element->next;
        DisposePtr((Ptr) element);
        element = next;
    }
	
    *elements = NULL;
}

// disposes of a single device, closing and releaseing interface, freeing memory fro device and elements, setting device pointer to NULL
// all your device no longer belong to us... (i.e., you do not 'own' the device anymore)
recDevice * HIDDisposeDevice(recDevice ** device)
{
    kern_return_t result = KERN_SUCCESS;
    recDevice * next = NULL;
	
    if (*device)
	{
        // save next device prior to disposing of this device
        next = (*device)->next;

        // free element lists
        HIDDisposeElementList(&(*device)->firstAxis);
        HIDDisposeElementList(&(*device)->firstButton);
        HIDDisposeElementList(&(*device)->firstHat);

        result = HIDCloseReleaseInterface(*device);   /* function sanity checks interface value (now application does not own device) */
        if (result != kIOReturnSuccess)
		{
            //TRACE0("HIDCloseReleaseInterface failed when trying to dipose device.\n")
		}
        
		DisposePtr((Ptr) * device);
        *device = NULL;
    }
	
    return next;
}

#endif //__APPLE__