/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "input-darwin.h"

#if defined(__APPLE__)

namespace hydra
{
// Gamepad thresholds
//
#define LEFT_THUMB_DEADZONE  7849
#define RIGHT_THUMB_DEADZONE 8689
#define TRIGGER_THRESHOLD    30

// Desc: Converts SHORT thumbstick values to FLOAT, while enforcing a deadzone
static inline float convertThumbstickValue(long thumbstickValue, short deadZone)
{
	if (thumbstickValue > +deadZone)
	{
		return (thumbstickValue - deadZone) / (32767.0f - deadZone);
	}
	if (thumbstickValue < -deadZone)
	{
		return (thumbstickValue + deadZone + 1.0f) / (32767.0f - deadZone);
	}

	return 0.0f;
}

IMPLEMENT_MODULE(inputDarwin, usr/lib/input)

inputDarwin::inputDarwin()
{
	m_device = NULL;
}
inputDarwin::~inputDarwin()
{
	while (m_device != NULL)
        m_device = HIDDisposeDevice(&m_device);
		
	m_device = NULL;
}

//
bool inputDarwin::open(HWND hwnd)
{
	// get master port, for accessing I/O Kit
	mach_port_t	masterPort = 0;
    IOReturn result = IOMasterPort(MACH_PORT_NULL, &masterPort);
	if (result != kIOReturnSuccess)
		return true;
		
    // set up a matching dictionary to search I/O Registry by class name for all HID class devices.
    CFMutableDictionaryRef hidDictionary = IOServiceMatching(kIOHIDDeviceKey);
	
	// now search I/O Registry for matching devices.
	io_iterator_t iterator = 0;
    result = IOServiceGetMatchingServices(masterPort, hidDictionary, &iterator);
	if (result != kIOReturnSuccess)
		return true;
	
	io_object_t hidDevice = 0;
	
	// build flat linked list of devices from device iterator
	m_device = NULL;

	recDevice * lastDevice = NULL;
	recDevice * device = NULL;
	while (hidDevice = IOIteratorNext(iterator))
	{
		device = HIDBuildDevice(hidDevice);
		if (!device)
			continue;
			
		// dump device object, it is no longer needed
        result = IOObjectRelease(hidDevice);
		
		// Filter device list to non-keyboard/mouse stuff
        if ((device->usagePage != kHIDPage_GenericDesktop) ||
            ((device->usage != kHIDUsage_GD_Joystick &&
              device->usage != kHIDUsage_GD_GamePad &&
              device->usage != kHIDUsage_GD_MultiAxisController)))
		{

            // release memory for the device
            HIDDisposeDevice(&device);
            DisposePtr((Ptr) device);
            continue;
        }

        // add device to the end of the list
        if (lastDevice)
            lastDevice->next = device;
        else
            m_device = device;

        lastDevice = device;
	}
	IOObjectRelease(iterator);

	// count the total number of devices we found
	int count = 0;
    device = m_device;
    while (device)
	{
        count ++;
		TRACE("%s\n", device->product)
        device = device->next;
    }
	
	TRACE(">> user: %d\n", count)
	
	return true;
}

void inputDarwin::update()
{
	int n = 0;

	recElement *element;
	SInt32 value;
	int i;

    recDevice * device = m_device;
	while (device)
	{
		// No device, set all inputs to zero
		m_padInfo[n].x[0] = m_padInfo[n].x[1] = m_padInfo[n].x[2] = 0;
		m_padInfo[n].y[0] = m_padInfo[n].y[1] = m_padInfo[n].y[2] = 0;
		m_padInfo[n].z[0] = m_padInfo[n].z[1] = m_padInfo[n].z[2] = 0;

		m_padInfo[n].gyro = 0;
		m_padInfo[n].buttons = 0;

		// device was unplugged; ignore it.
		if (device->removed)
		{      
			if (device->uncentered)
				device->uncentered = 0;

			return;
		}

		i = 0;
		element = device->firstAxis;
		while (element)
		{
			value = HIDScaledCalibratedValue(device, element, -32768, 32768);
			switch (i)
			{
			case 0:
				m_padInfo[n].x[0] = convertThumbstickValue(value, LEFT_THUMB_DEADZONE);
				break;
			case 1:
				m_padInfo[n].y[0] = -convertThumbstickValue(value, LEFT_THUMB_DEADZONE);
				break;
			case 2:
				m_padInfo[n].x[1] = convertThumbstickValue(value, RIGHT_THUMB_DEADZONE);
				break;
			case 3:
				m_padInfo[n].y[1] = -convertThumbstickValue(value, RIGHT_THUMB_DEADZONE);	
				break;
			case 4:
				m_padInfo[n].x[2] = convertThumbstickValue(value, LEFT_THUMB_DEADZONE);
				break;
			case 5:
				m_padInfo[n].y[2] = convertThumbstickValue(value, LEFT_THUMB_DEADZONE);
				break;
			}

			element = element->next;
			++i;	
		}

		i = 0;
		element = device->firstButton;
		while (element)
		{
			value = HIDGetElementValue(device, element);
			if (value)
			{
				switch (i)
				{
				case 0:
					m_padInfo[n].buttons |= BUTTON_UP;
					break;
				case 1:
					m_padInfo[n].buttons |= BUTTON_DOWN;
					break;
				case 2:
					m_padInfo[n].buttons |= BUTTON_LEFT;
					break;
				case 3:
					m_padInfo[n].buttons |= BUTTON_RIGHT;
					break;

				case 4:
					m_padInfo[n].buttons |= BUTTON_START;
					break;
				case 5:
					m_padInfo[n].buttons |= BUTTON_BACK;
					break;
				case 6:
					m_padInfo[n].buttons |= BUTTON_L3;
					break;
				case 7:
					m_padInfo[n].buttons |= BUTTON_R3;
					break;
				case 8:
					m_padInfo[n].buttons |= BUTTON_L1;
					break;
				case 9:
					m_padInfo[n].buttons |= BUTTON_R1;
					break;
				case 10:
					m_padInfo[n].buttons |= BUTTON_HOME;
					break;
				case 11:
					m_padInfo[n].buttons |= BUTTON_A;
					break;
				case 12:
					m_padInfo[n].buttons |= BUTTON_B;
					break;
				case 13:
					m_padInfo[n].buttons |= BUTTON_X;
					break;
				case 14:
					m_padInfo[n].buttons |= BUTTON_Y;
					break;
				}
			}
			
			element = element->next;
			++i;
		}

        device = device->next;
		n ++;
	}
}

} //hydra

#endif //__APPLE__
