/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "input-dinput8.h"

#if defined(_WIN32)

namespace hydra
{

static BOOL CALLBACK enumJoysticksCallback(LPCDIDEVICEINSTANCE pdidInstance, void * param);
static BOOL CALLBACK enumObjectsCallback(LPCDIDEVICEOBJECTINSTANCE pdidoi, void * param);

#define THUMB_DEADZONE 239
static inline float convertThumbstickValue(LONG thumbstickValue, short deadZone)
{
	if (thumbstickValue > +deadZone)
	{
		return (thumbstickValue - deadZone) / (1000.0f - deadZone);
	}
	if (thumbstickValue < -deadZone)
	{
		return (thumbstickValue + deadZone) / (1000.0f - deadZone);
	}

	return 0.0f;
}
static inline float convertTriggerValue(LONG thumbstickValue, short deadZone)
{
	if (thumbstickValue > +deadZone)
	{
		return (thumbstickValue - deadZone + 1.0f) / (1000.0f - deadZone);
	}
	if (thumbstickValue < -deadZone)
	{
		return (thumbstickValue + deadZone) / (1000.0f - deadZone);
	}

	return 0.0f;
}

struct tagCALLBACK
{
	HWND hwnd;
	LPDIRECTINPUT8 input;
	LPDIRECTINPUTDEVICE8 device;
};

static unsigned char padStatus[USER_MAX_COUNT];		// old pad status

IMPLEMENT_MODULE(inputDInput8, usr/lib/input)

inputDInput8::inputDInput8()
{
	m_input = NULL;
	m_device = NULL;

	//
	m_initialize = false;
}

inputDInput8::~inputDInput8()
{
	if (m_input)
	{
		m_input->Release();
		m_input = NULL;
	}

	if (m_device)
	{
		m_device->Release();
		m_device = NULL;
	}
}

bool inputDInput8::open(HWND hwnd)
{
	//
	for (int n = 0; n < USER_MAX_COUNT; n ++)
		padStatus[n] = 0;

	if (FAILED(DirectInput8Create(GetModuleHandle(NULL), 
		DIRECTINPUT_VERSION,
		IID_IDirectInput8,
		(VOID **) &m_input, 
		NULL)))
		return false;

	//
	initialize(hwnd);

	return true;
}

bool inputDInput8::initialize(HWND hwnd)
{
	tagCALLBACK callBack;
	callBack.hwnd = hwnd;
	callBack.input = m_input;
	callBack.device = NULL;

	if (FAILED(m_input->EnumDevices(DI8DEVCLASS_GAMECTRL, 
		enumJoysticksCallback,
		(VOID *) &callBack, 
		DIEDFL_ATTACHEDONLY)))
		return false;

	m_device = callBack.device;
	if (!m_device)
		return false;

	if (FAILED(m_device->SetDataFormat(&c_dfDIJoystick2)))
		return false;

	if (FAILED(m_device->SetCooperativeLevel(hwnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND)))
		return false;

	if (FAILED(m_device->EnumObjects(enumObjectsCallback, (void *) &callBack, DIDFT_ALL)))
		return false;

	//
	m_initialize = true;

	return true;
}

void inputDInput8::update()
{
	//if (!m_initialize)
	//	init();

	//
	if (!m_initialize)
		return;

	if (FAILED(m_device->Poll()))
	{
		HRESULT result = m_device->Acquire();
		while (result == DIERR_INPUTLOST)
		{
			result = m_device->Acquire();
		};
	}

	int n = 0;

	// No device, set all inputs to zero
	m_padInfo[n].x[0] = m_padInfo[n].x[1] = m_padInfo[n].x[2] = 0;
	m_padInfo[n].y[0] = m_padInfo[n].y[1] = m_padInfo[n].y[2] = 0;
	m_padInfo[n].z[0] = m_padInfo[n].z[1] = m_padInfo[n].z[2] = 0;

	m_padInfo[n].gyro = 0;
	m_padInfo[n].buttons = 0;

	DIJOYSTATE2 diJoyState;
	if (FAILED(m_device->GetDeviceState(sizeof(DIJOYSTATE2), &diJoyState)))
		return;

	switch (diJoyState.rgdwPOV[0])
	{
	case 0:
		m_padInfo[n].buttons |= BUTTON_UP;
		break;

	case 9000:
		m_padInfo[n].buttons |= BUTTON_RIGHT;
		break;

	case 18000:
		m_padInfo[n].buttons |= BUTTON_DOWN;
		break;

	case 27000:
		m_padInfo[n].buttons |= BUTTON_LEFT;
		break;
	}

	// button
	for (int btn = 0; btn < 18; btn++)
	{
		if (diJoyState.rgbButtons[btn] & 0x80)
		{
			switch (btn)
			{
			case 0:
				m_padInfo[n].buttons |= BUTTON_A;
				break;

			case 1:
				m_padInfo[n].buttons |= BUTTON_B;
				break;

			case 2:
				m_padInfo[n].buttons |= BUTTON_X;
				break;

			case 3:
				m_padInfo[n].buttons |= BUTTON_Y;
				break;

			case 4:
				m_padInfo[n].buttons |= BUTTON_L1;
				break;

			case 5:
				m_padInfo[n].buttons |= BUTTON_R1;
				break;

			case 6:
				m_padInfo[n].buttons |= BUTTON_BACK;
				break;

			case 7:
				m_padInfo[n].buttons |= BUTTON_START;
				break;

			case 8:
				m_padInfo[n].buttons |= BUTTON_L3;
				break;

			case 9:
				m_padInfo[n].buttons |= BUTTON_R3;
				break;
			}
		}
	}

	m_padInfo[n].x[0] = convertThumbstickValue(diJoyState.lX, THUMB_DEADZONE);
	m_padInfo[n].y[0] = -convertThumbstickValue(diJoyState.lY, THUMB_DEADZONE);;

	m_padInfo[n].x[1] = convertThumbstickValue(diJoyState.lRx, THUMB_DEADZONE);
	m_padInfo[n].y[1] = -convertThumbstickValue(diJoyState.lRy, THUMB_DEADZONE);

	m_padInfo[n].x[2] = convertTriggerValue(diJoyState.lZ, THUMB_DEADZONE);
	m_padInfo[n].y[2] = -convertThumbstickValue(diJoyState.lRz, THUMB_DEADZONE);;
}

static BOOL CALLBACK enumJoysticksCallback(LPCDIDEVICEINSTANCE pdidInstance, void * param)
{
	struct tagCALLBACK * callBack = (tagCALLBACK *) param;

	// Obtain an interface to the enumerated joystick.
	LPDIRECTINPUT8 input = callBack->input;

	LPDIRECTINPUTDEVICE8 device;
	if (FAILED(input->CreateDevice(pdidInstance->guidInstance, &device, NULL)))
		// If it failed, then we can't use this joystick. (Maybe the user unplugged
		// it while we were in the middle of enumerating it.)
		return DIENUM_CONTINUE;

	callBack->device = device;

	// Stop enumeration. Note: we're just taking the first joystick we get. You
	// could store all the enumerated joysticks and let the user pick.
	return DIENUM_STOP;
}
static BOOL CALLBACK enumObjectsCallback(LPCDIDEVICEOBJECTINSTANCE pdidoi, void * param)
{
	struct tagCALLBACK * callBack = (tagCALLBACK *) param;

	static int sliderCount = 0;  // Number of returned slider controls
	static int povCount = 0;     // Number of returned POV controls

	// For axes that are returned, set the DIPROP_RANGE property for the
	// enumerated axis in order to scale min/max values.
	if (pdidoi->dwType & DIDFT_AXIS)
	{
		DIPROPRANGE diprg; 
		diprg.diph.dwSize       = sizeof(DIPROPRANGE); 
		diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
		diprg.diph.dwHow        = DIPH_BYID; 
		diprg.diph.dwObj        = pdidoi->dwType; // Specify the enumerated axis
		diprg.lMin              = -1000; 
		diprg.lMax              = +1000; 

		// Set the range for the axis
		LPDIRECTINPUTDEVICE8 device = callBack->device;
		if (FAILED(device->SetProperty(DIPROP_RANGE, &diprg.diph))) 
			return DIENUM_STOP;
	}

	return DIENUM_CONTINUE;
}

} //hydra

#endif //_WIN32
