/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "window-carbon.h"

#include "graphics/igraphics.h"

#include "kernel.h"

#if defined(__APPLE__)

namespace hydra
{

static OSStatus appEventHandler(EventHandlerCallRef inCaller, EventRef inEvent, void * inRefcon);
static OSStatus windowEventHandler(EventHandlerCallRef inCaller, EventRef inEvent, void * inRefcon);

static void renderTimedEventProcessor(EventLoopTimerRef inTimer, void * timeData);

static EventLoopTimerRef renderTimer = NULL;
static EventLoopTimerUPP renderTimerUPP = NULL;

static EventHandlerUPP appEventProcessor = NULL;
static EventHandlerUPP windowEventProcessor = NULL;

IMPLEMENT_MODULE(windowCarbon, usr/lib/window)

windowCarbon::windowCarbon()
{
}
windowCarbon::~windowCarbon()
{
	// remove and dispose of our timers
	RemoveEventLoopTimer(renderTimer);
	DisposeEventLoopTimerUPP(renderTimerUPP);
	
	// remove our other handlers
	DisposeEventHandlerUPP(appEventProcessor);
	DisposeEventHandlerUPP(windowEventProcessor);

	DisposeWindow(m_hwnd);
}

//
bool windowCarbon::create(int width, int height, const char * title)
{
	Rect rect = { 0, 0, width, height };
	OSStatus result = CreateNewWindow(kDocumentWindowClass, 
			  kWindowStandardFloatingAttributes |
			  //kWindowNoTitleBarAttribute |
			  kWindowStandardHandlerAttribute,
			  &rect, 
			  &m_hwnd);		
	SetWindowTitleWithCFString(m_hwnd, CFSTR(PLATFORMNAME " for " PLATFORM));

    // position new windows in a staggered arrangement on the main screen
    RepositionWindow(m_hwnd, NULL, kWindowCascadeOnMainScreen);

    // the window was created hidden, so show it
    ShowWindow(m_hwnd);

	//
	static const EventTypeSpec kAppEvents[] =
    {
        { kEventClassCommand, kEventCommandProcess }
    };

    // install our handler for common commands on the application target
	appEventProcessor = NewEventHandlerUPP(appEventHandler);
    InstallApplicationEventHandler(appEventProcessor,
                                   GetEventTypeCount(kAppEvents),
								   kAppEvents,
                                   0,
								   NULL);

    static const EventTypeSpec kWindowEvents[] =
    {
        { kEventClassCommand, kEventCommandProcess },
		
		{ kEventClassWindow, kEventWindowClose },
		{ kEventClassWindow, kEventWindowDrawContent },
		{ kEventClassWindow, kEventWindowBoundsChanged },
		{ kEventClassWindow, kEventWindowHitTest },
    };
	
    // install a command handler on the window. We don't use this handler yet, but nearly all
    // Carbon apps will need to handle commands, so this saves everyone a little typing.
	windowEventProcessor = NewEventHandlerUPP(windowEventHandler);
    InstallWindowEventHandler(m_hwnd,
	                          windowEventProcessor,
                              GetEventTypeCount(kWindowEvents),
							  kWindowEvents,
                              m_hwnd,
							  NULL);

	// render Timed event handler installer
	renderTimerUPP = NewEventLoopTimerUPP(renderTimedEventProcessor);
    result = InstallEventLoopTimer(GetCurrentEventLoop(),
				0,
				kEventDurationMillisecond,
				renderTimerUPP,
				NULL, // User Data
				&renderTimer);

	return true;
}

//
int windowCarbon::run()
{
    // run the event loop
    RunApplicationEventLoop();
	
	return 0;
}

static void renderTimedEventProcessor(EventLoopTimerRef inTimer, void * timeData)
{
    #pragma unused(inTimer, timeData)
	
    // call the main rendering loop here
    windowCarbon::idle();
}

//
static OSStatus appEventHandler(EventHandlerCallRef inCaller, EventRef inEvent, void * inRefcon)
{
    OSStatus result = eventNotHandledErr;
    
    switch (GetEventClass(inEvent))
    {
    case kEventClassCommand:
        {
        HICommandExtended cmd;
        verify_noerr(GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof(cmd), NULL, &cmd));
            
        switch (GetEventKind(inEvent))
        {
        case kEventCommandProcess:
			switch (cmd.commandID) 
            {
            case kHICommandNew:
				break;
                            
                // Add your own command-handling cases here
                        
            default:
                break;
            }
            break;
		}
        break;
        }
            
    default:
        break;
    }
    
    return result;
}

static OSStatus windowEventHandler(EventHandlerCallRef inCaller, EventRef inEvent, void * inRefcon)
{
    OSStatus result = eventNotHandledErr;
    
	WindowRef hwnd;
	Rect bounds;
	
    switch (GetEventClass(inEvent))
    {
    case kEventClassCommand:
		{
        HICommandExtended cmd;
        verify_noerr(GetEventParameter(inEvent,
						 kEventParamDirectObject,
						 typeHICommand,
						 NULL,
						 sizeof(cmd),
						 NULL,
						 &cmd));
            
        switch (GetEventKind(inEvent))
        {
		case kEventCommandProcess:
			switch (cmd.commandID)
            {
            // Add your own command-handling cases here
            default:
				break;
            }
            break;
        }
        break;
		}	

	case kEventClassWindow:		// 'wind'
		GetEventParameter(inEvent,		// the event itself
			kEventParamDirectObject, 	// symbolic parameter name
			typeWindowRef,				// expected type 
			NULL,						// actual type (NULL is valid)
			sizeof(hwnd),				// buffer size
			NULL,						// actual buffer size (Can be NULL)
			&hwnd);						// variable to hold data

		switch(GetEventKind(inEvent))
		{
		case kEventWindowDrawContent:
            // handle window updates here
            result = noErr;
            break;
		case kEventWindowHitTest:
			break;

		case kEventWindowBoundsChanging:
        case kEventWindowBoundsChanged:
			{
			UInt32 attributes;
			GetEventParameter(inEvent,
				kEventParamAttributes,
				typeUInt32,
				NULL,
				sizeof(UInt32),
				NULL,
				&attributes);

			if (attributes & kWindowBoundsChangeSizeChanged)
			{
				GetEventParameter(inEvent, kEventParamCurrentBounds, typeQDRectangle, NULL, sizeof(bounds), NULL, &bounds);
				InvalWindowRect(hwnd, GetWindowPortBounds(hwnd, &bounds));
			  
				if (IGraphics::pmodule != NULL)
					graphicsModule()->setViewport(0, 0, bounds.right - bounds.left, bounds.bottom - bounds.top);
				
				//TRACE(">> kWindowBoundsChangeSizeChanged(%d, %d)\n", bounds.right - bounds.left, bounds.bottom - bounds.top)
			}

            result = noErr;
            break;
			}	
        case kEventWindowClose:
            DisposeEventHandlerUPP(windowEventProcessor);
            DisposeWindow(hwnd);
			
			QuitApplicationEventLoop();		
            result = noErr;
            break;
        default:
            break;
		}
		break;

    default:
        break;
    }
    
    return result;
}

} //hydra

#endif //__APPLE__
