/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */

#include "camera.h"

namespace hydra
{

Camera::Camera()
{
	m_view._11 = 1;
	m_view._12 = 0;
	m_view._13 = 0;
	m_view._14 = 0;

	m_view._21 = 0;
	m_view._22 = 1;
	m_view._23 = 0;
	m_view._24 = 0;
	
	m_view._31 = 0;
	m_view._32 = 0;
	m_view._33 = -1;
	m_view._34 = 0;
	
	m_view._41 = 0;
	m_view._42 = 0;
	m_view._43 = 0;
	m_view._44 = 1;
}

void Camera::setPos(const float3 & pos)
{
	m_view._41 = pos.x;
	m_view._42 = pos.y;
	m_view._43 = pos.z;
}

void Camera::setRotationYawPitchRoll(const float yaw, const float pitch, const float roll)
{
	float cosx = cosf(pitch);
	float cosy = cosf(yaw);
	float cosz = cosf(roll);

	float sinx = sinf(pitch);
	float siny = sinf(yaw);
	float sinz = sinf(roll);

	m_view._11 =  cosy * cosz + sinx * siny * sinz;
	m_view._12 = -cosx * sinz;
	m_view._13 =  sinx * cosy * sinz - siny * cosz;

	m_view._21 =  cosy * sinz - sinx * siny * cosz;
	m_view._22 =  cosx * cosz;
	m_view._23 = -siny * sinz - sinx * cosy * cosz;

	m_view._31 =  cosx * siny;
	m_view._32 =  sinx;
	m_view._33 =  cosx * cosy;
}

void Camera::setLookAt(const float3 & eye, const float3 & lookAt, const float3 & up)
{
	// normalizing the viewing vector 
	float3 v = normalize(lookAt - eye);

	// normalising the up vector. no need for this here if you have your 
	// up vector already normalised, which is mostly the case. 
	float3 u = normalize(up);

	float3 right = crossProduct(v, u); 
	//u = crossProduct(right, v); 

#if defined(_OGLES)
	// Here's what the final view matrix should look like:
	//
	//  |  rx   ry   rz  -(r.e) |
	//  |  ux   uy   uz  -(u.e) |
	//  | -lx  -ly  -lz   (l.e) |
	//  |   0    0    0     1   |
	//
	// Where r = Right vector
	//       u = Up vector
	//       l = Look vector
	//       e = Eye position in world space
	//       . = Dot-product operation

	m_view._11 = right.x;
	m_view._12 = u.x;
	m_view._13 = -v.x;
	m_view._14 = 0;

	m_view._21 = right.y;
	m_view._22 = u.y;
	m_view._23 = -v.y;
	m_view._24 = 0;

	m_view._31 = right.z;
	m_view._32 = u.z;
	m_view._33 = -v.z;
	m_view._34 = 0;

	m_view._41 = dotProduct(right, eye);
	m_view._42 = dotProduct(u, eye);
	m_view._43 = dotProduct(v, eye);
	m_view._44 = 1;

#elif defined(_PSP)
	sceGumMatrixMode(SCEGU_MATRIX_WORLD);
	sceGumLookAt((ScePspFVector3 *) &eye,
		(ScePspFVector3 *) &lookAt,
		(ScePspFVector3 *) &up);

#else
	// Here's what the final matrix should look like:
	//
	//  |   rx     ux     lx    0 |
	//  |   ry     uy     ly    0 |
	//  |   rz     uz     lz    0 |
	//  | -(r.e) -(u.e) -(l.e)  1 |
	//
	// Where r = Right vector
	//       u = Up vector
	//       l = Look vector
	//       e = Eye position in world space
	//       . = Dot-product operation

	m_view._11 = right.x;
	m_view._12 = u.x;
	m_view._13 = v.x;
	m_view._14 = 0;

	m_view._21 = right.y;
	m_view._22 = u.y;
	m_view._23 = v.y;
	m_view._24 = 0;

	m_view._31 = right.z;
	m_view._32 = u.z;
	m_view._33 = v.z;
	m_view._34 = 0;

	m_view._41 = -dotProduct(right, eye);
	m_view._42 = -dotProduct(u, eye);
	m_view._43 = -dotProduct(v, eye);
	m_view._44 = 1;
#endif //_OGLES
}

void Camera::setPerspectiveFov(const float fovy, const float aspect, const float zNear, const float zFar)
{
#if defined(_OGLES)
	// This code is based off the MESA source for gluPerspective
	// *NOTE* This assumes GL_PROJECTION is the current matrix
	float xmin, xmax, ymin, ymax;

	xmax = zNear * (float) tan(fovy * M_PI / 360.0f);
	xmin = -xmax;

	ymin = xmin / aspect;
	ymax = xmax / aspect;

	// Set up the projection matrix
	m_projection._11 = (2.0f * zNear) / (xmax - xmin);
	m_projection._12 = 0;
	m_projection._13 = 0;
	m_projection._14 = 0;

	m_projection._21 = 0;
	m_projection._22 = (2.0f * zNear) / (ymax - ymin);
	m_projection._23 = 0;
	m_projection._24 = 0;

	m_projection._31 = (xmax + xmin) / (xmax - xmin);
	m_projection._32 = (ymax + ymin) / (ymax - ymin);
	m_projection._33 = -(zFar + zNear) / (zFar - zNear);
	m_projection._34 = -1.0f; 

	m_projection._41 = 0;
	m_projection._42 = 0;
	m_projection._43 = -(2.0f * zFar * zNear) / (zFar - zNear); 
	m_projection._44 = 0;

#elif defined(_PSP)
	sceGumMatrixMode(SCEGU_MATRIX_PROJECTION);
	sceGumPerspective(40.0f * 3.14f / 180.0f, SCEGU_SCR_ASPECT, zNear, zFar);

#else
	float h = float(1.0f / tan(TORADIAN(fovy) * 0.5f));
	float w = h / aspect;

	// Set up the projection matrix
	m_projection._11 = w;
	m_projection._12 = 0;
	m_projection._13 = 0;
	m_projection._14 = 0;

	m_projection._21 = 0;
	m_projection._22 = h;
	m_projection._23 = 0;
	m_projection._24 = 0;

	m_projection._31 = 0;
	m_projection._32 = 0;
	m_projection._33 = zFar / (zFar - zNear);
	m_projection._34 = 1.0f;

	m_projection._41 = 0;
	m_projection._42 = 0;
	m_projection._43 = -zNear * (zFar / (zFar - zNear)); 
	m_projection._44 = 0;
#endif //_OGLES
}

} //hydra
